"""RTR Batch GET abstraction module."""
from __future__ import annotations

import os

from dataclasses import dataclass
from typing import TYPE_CHECKING

import py7zr


if TYPE_CHECKING:
    # This trick will avoid us from causing a cyclical reference within the class
    # Credit: https://stackoverflow.com/a/39757388
    from caracara.modules.rtr.batch_session import RTRBatchSession


@dataclass
class GetFile:
    """
    Represents a file uploaded to Falcon via a GET command.

    This class may be instantiated many times, with each object stored in a common list,
    in order to represent many files retrieved from a GET comamnd executed against a
    batch session.
    """

    device_id: str = None
    filename: str = None
    session_id: str = None
    sha256: str = None
    size: int = None
    batch_session: RTRBatchSession = None

    def download(
        self,
        output_path: str,
        extract: bool = True,
        preserve_7z: bool = False,
    ):
        """
        Download a file to a specified path.

        If the path is a folder, the filename will be auto generated.
        If the path is a file, it'll be downloaded to that path and name.
        """
        if self.filename.startswith("/"):
            # macOS or *nix path
            filename = self.filename.rsplit("/", maxsplit=1)[-1]
        else:
            # Windows filename
            filename = self.filename.rsplit("\\", maxsplit=1)[-1]

        filename_noext, ext = os.path.splitext(filename)

        if os.path.isdir(output_path):
            # Output path is a folder, so we should compute the filename
            full_output_path = os.path.join(
                output_path,
                f"{filename_noext}_{self.sha256}_{self.device_id}{ext}",
            )
            full_output_path_7z = full_output_path + ".7z"
        else:
            full_output_path = output_path
            if extract:
                full_output_path_7z = full_output_path + ".7z"
            else:
                # We should ensure that the user's path ends in .7z if they are not extracting
                if not full_output_path.endswith(".7z"):
                    full_output_path_7z = full_output_path + ".7z"

        file_contents = self.batch_session.api.get_extracted_file_contents(
            session_id=self.session_id,
            sha256=self.sha256,
            filename=self.filename,
        )

        with open(full_output_path_7z, 'wb') as output_7z_file:
            output_7z_file.write(file_contents)

        if not extract:
            # Downloaded, so we're done now!
            return

        with py7zr.SevenZipFile(  # nosec - The password 'infected' is generic and always the same
            file=full_output_path_7z,
            mode='r',
            password='infected',
        ) as archive:
            inner_filename = archive.getnames()[0]
            target_dir = os.path.dirname(full_output_path_7z)

            archive.extract(target_dir, inner_filename)
            os.rename(
                os.path.join(target_dir, inner_filename),
                full_output_path,
            )

            if not preserve_7z:
                os.unlink(full_output_path_7z)

    def __str__(self):
        """Return a loggable string representing the contents of the object."""
        return f"Filename {self.filename} | Session ID: {self.session_id} | SHA256: {self.sha256}"



"""Falcon Flight Control (MSSP) APIs.

This module handles interaction with Parent CIDs, which have beneath them children (members) that
can be interfaced with directly.

This module handles interactions with the Parent CID's management APIs. To authenticate to
a child/member CID, use the member_cid kwarg when initialising a Caracara Client object.
"""
from typing import (
    Dict,
    List,
    Union,
)

from falconpy import (
    FlightControl,
    OAuth2,
)

from caracara.common.batching import batch_get_data
from caracara.common.module import FalconApiModule
from caracara.common.pagination import all_pages_numbered_offset_parallel


class FlightControlApiModule(FalconApiModule):
    """The FlightControlApiModule facilitates interactions with the Flight Control (MSSP) API."""

    name = "CrowdStrike Falcon Flight Control API Module"
    help = "Interact with the management API calls in a Falcon Flight Control (MSSP) Parent CID"

    def __init__(self, api_authentication: OAuth2):
        """Construct an instance of the FlightControlApiModule class."""
        super().__init__(api_authentication)

        self.logger.debug("Configuring the FalconPy Flight Control API")
        self.flight_control_api = FlightControl(auth_object=self.api_authentication)

    def get_child_cids(self) -> List[str]:
        """Retreives a list of every Child/Member CID within a Parent instance.

        This function will return a list of CID strings without additional data. To get the names
        of the CIDs, either pass the results of this function into get_child_cid_data(), or use the
        describe_child_cids() function to obtain all this data in one function call.

        Returns
        -------
        list: A list of all Child CIDs owned by the Parent CID associated with this Caracara Client
              instance
        """
        self.logger.info("Obtaining a list of all Child CIDs")
        child_cids: List[str] = all_pages_numbered_offset_parallel(
            self.flight_control_api.query_children,
            self.logger,
            10,
        )
        return child_cids

    def get_child_cid_data(self, cids: List[str]) -> Dict[str, Dict[str, Union[List[str], str]]]:
        """Return a dictionary containing details for every Child CID specified by ID.

        You should use this endpoint if you have a list of Child CIDs associated with the Parent
        CID you are authenticated to within this Caracara Client instance. If you do not yet have
        a list of Child CIDs, either use the get_child_cids() function to obtain these first,
        or alternatively use the describe_child_cids() function to obtain all this data in one
        function call.

        Arguments
        ---------
        cids: [str], required
            A list of Falcon Customer IDs (CIDs) associated with this Parent Falcon tenant

        Returns
        -------
        dict: A dictionary containing details for every Child CID listed.
        """
        self.logger.info("Obtaining data for %d child CIDs", len(cids))
        child_cid_data: Dict[str, Dict[str, Union[List[str], str]]] = batch_get_data(
            cids,
            self.flight_control_api.get_children,
        )
        return child_cid_data

    def describe_child_cids(self) -> Dict[str, Dict[str, Union[List[str], str]]]:
        """Return a dictionary containing details for each Child CID associated with this tenant.

        Returns
        -------
        dict: A dictionary containing details for every Child CID associated with this
              Parent Falcon tenant (CID)
        """
        self.logger.info("Describing this Parent CID's Child CIDs")
        child_cids = self.get_child_cids()
        child_cid_data = self.get_child_cid_data(cids=child_cids)
        return child_cid_data



"""Caracara wrapper for Policies API.

This file contains wrapper classes that can represent policies in a generic way.
It is to be extended by the respective modules (response_policies, prevention_policies, etc.)
"""
# Disable TODO warning, to be removed when a wrapper is implemented for `ioa_rule_groups`
# pylint: disable=W0511
from __future__ import annotations
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, Dict, List


class PolicySetting(ABC):
    """Generic policy setting class.

    Contains a list of settings and settings groups. This is recursive, as Falcon allows
    nested policy settings.
    """

    def __init__(self, data_dict: Dict = None):
        """Initialise a policy setting from a Falcon dictionary."""
        self.name: str = None

        if data_dict:
            self._load_data_dict(data_dict)

    def _load_data_dict(self, data_dict: Dict):
        self.name = data_dict.get("name")

    @abstractmethod
    def dump(self) -> Dict:
        """Return a dictionary encompassing all Policy object content.

        This dictionary is compliant with the CrowdStrike API specification, and therefore
        can replicate the exact response sent back by the CrowdStrike API.
        """

    @abstractmethod
    def flat_dump(self) -> Dict:
        """Return a stripped down dictionary representing all settings and their values.

        This dictionary is designed for use with CrowdStrike API policy verbs (i.e., policy
        creation (POST) and modification (PATCH), and therefore is limited only to the content
        required to execute these actions.
        """


class GroupAssignment(PolicySetting):
    """Represents an assignment rule that maps a policy to a host group."""

    def __init__(self, data_dict: Dict = None):
        """Configure a new Group Assignment object."""
        self.assignment_rule: str = None
        self.created_by: str = None
        self.created_timestamp: datetime = None
        self.description: str = None
        self.group_id: str = None
        self.group_type: str = None
        self.modified_by: str = None
        self.modified_timestamp: datetime = None

        super().__init__(data_dict)

    def _load_data_dict(self, data_dict: Dict):
        self.assignment_rule = data_dict.get("assignment_rule")
        self.created_by = data_dict.get("created_by")
        self.created_timestamp = data_dict.get("created_timestamp")
        self.description = data_dict.get("description")
        self.group_id = data_dict.get("id")
        self.group_type = data_dict.get("group_type")
        self.modified_by = data_dict.get("modified_by")
        self.modified_timestamp = data_dict.get("modified_timestamp")

        super()._load_data_dict(data_dict)

    def dump(self) -> Dict:
        """Return a dictionary representing a group assignment.

        This dictionary is compliant with the CrowdStrike API specification, and therefore
        can replicate the exact response sent back by the CrowdStrike API.
        """
        return {
            "assignment_rule": self.assignment_rule,
            "created_by": self.created_by,
            "created_timestamp": self.created_timestamp,
            "description": self.description,
            "group_type": self.group_type,
            "id": self.group_id,
            "modified_by": self.modified_by,
            "modified_timestamp": self.modified_timestamp,
            "name": self.name,
        }

    def flat_dump(self) -> Dict:
        """Return a stripped down dictionary representing a group assignment.

        This dictionary is designed for use with CrowdStrike API policy verbs (i.e., policy
        creation (POST) and modification (PATCH), and therefore is limited only to the content
        required to execute these actions.
        """
        return {
            "assignment_rule": self.assignment_rule,
            "description": self.description,
            "group_type": self.group_type,
            "id": self.group_id,
            "name": self.name,
        }


class ChangeablePolicySetting(PolicySetting, ABC):
    """Policy setting with an attribbute that can be changed.

    The actual type of setting must be derived from this (e.g., a toggle setting).
    """

    def __init__(self, data_dict: Dict = None):
        """Configure a new Changeable Policy Setting.

        This constructor only handles parameters that are available in all policy settings.
        Individual policy settings (such as toggles and sliders) have their own additional
        settings to store the values that are passed to and from the API.
        """
        self.description: str = None
        self.name: str = None
        self.setting_id: str = None

        super().__init__(data_dict)

    @property
    @abstractmethod
    def setting_type(self) -> str:
        """Store the type of setting.

        Examples: toggle, mlslider.
        """

    @abstractmethod
    def _dump_value(self) -> Dict[str, str]:
        """Generate the setting type-specific value dictionary."""

    def _load_data_dict(self, data_dict: Dict):
        super()._load_data_dict(data_dict)

        self.description = data_dict.get("description")
        self.setting_id = data_dict.get("id")
        self.setting_type = data_dict.get("type")

    def dump(self) -> Dict:
        """Return a dictionary representing a policy setting.

        This dictionary is compliant with the CrowdStrike API specification, and therefore
        can replicate the exact response sent back by the CrowdStrike API if this policy
        already happened to exist in the Cloud.
        """
        return {
            "description": self.description,
            "id": self.setting_id,
            "name": self.name,
            "type": self.setting_type,
            "value": self._dump_value(),
        }

    def flat_dump(self) -> Dict:
        """Return a stripped down dictionary representing a policy setting.

        This dictionary is designed for use with CrowdStrike API policy verbs (i.e., policy
        creation (POST) and modification (PATCH), and therefore is limited only to the content
        required to execute these actions.
        """
        return {
            "id": self.setting_id,
            "value": self._dump_value(),
        }


class TogglePolicySetting(ChangeablePolicySetting):
    """Toggle policy setting that has two options: enabled and disabled."""

    setting_type = "toggle"

    def __init__(self, data_dict: Dict = None):
        """Configure a new Toggle.

        We default this to None so that no position is ever assumed.
        """
        self.enabled: bool = None

        super().__init__(data_dict)

    def _dump_value(self) -> Dict[str, str]:
        return {
            "enabled": self.enabled,
        }

    def _load_data_dict(self, data_dict: Dict):
        super()._load_data_dict(data_dict)

        value_dict = data_dict.get("value")
        self.enabled = value_dict.get("enabled")


class MLSliderPolicySetting(ChangeablePolicySetting):
    """Machine Learning (ML) policy setting.

    Each ML slider has string values for detection and prevention.
    """

    setting_type = "mlslider"

    def __init__(self, data_dict: Dict = None):
        """Configure a new ML Slider.

        We initially set detection and prevention to None so that no assumption is made
        about the protection levels stored by the Cloud or desired by a user.
        """
        self.detection: str = None
        self.prevention: str = None

        super().__init__(data_dict)

    def _dump_value(self) -> Dict[str, str]:
        return {
            "detection": self.detection,
            "prevention": self.prevention,
        }

    def _load_data_dict(self, data_dict: Dict):
        super()._load_data_dict(data_dict)

        value_dict: Dict = data_dict.get("value")
        self.detection = value_dict.get("detection")
        self.prevention = value_dict.get("prevention")


SETTINGS_TYPE_MAP = {
    "mlslider": MLSliderPolicySetting,
    "toggle": TogglePolicySetting,
}


class PolicySettingGroup(PolicySetting):
    """Policy setting with a list of other settings within it."""

    def __init__(self, data_dict: Dict = None):
        """Return a new policy settings group, optionally configured via a dictionary."""
        self.settings: List[PolicySetting] = []

        super().__init__(data_dict)

    def _load_data_dict(self, data_dict: Dict):
        super()._load_data_dict(data_dict)

        inner_settings: List[Dict] = data_dict.get("settings", [])
        for inner_setting_dict in inner_settings:
            setting_template_name = inner_setting_dict.get("type")
            if setting_template_name in SETTINGS_TYPE_MAP:
                setting_template = SETTINGS_TYPE_MAP[setting_template_name]
                self.settings.append(setting_template(data_dict=inner_setting_dict))
            else:
                raise Exception(f"Setting type {setting_template_name} is not yet supported")

    def dump(self) -> Dict:
        """Return a dictionary representing a policy settings group.

        This dictionary is compliant with the CrowdStrike API specification, and therefore
        can replicate the exact response sent back by the CrowdStrike API if this policy
        already happened to exist in the Cloud.
        """
        return {
            "name": self.name,
            "settings": [x.dump() for x in self.settings],
        }

    def flat_dump(self) -> Dict:
        """Return a stripped down dictionary representing a policy settings group.

        This dictionary is designed for use with CrowdStrike API policy verbs (i.e., policy
        creation (POST) and modification (PATCH), and therefore is limited only to the content
        required to execute these actions.
        """
        return {
            "settings": [x.flat_dump() for x in self.settings],
        }


class Policy:
    """A generic policy class.

    Each policy object is comprised of one or more policy settings groups. These groups in turn
    contain individual policy settings that correspond to settings available within the Falcon
    Console. Each group of settings is handled by a PolicySettingGroup object.

    Each policy can be assigned to one or more groups. Each group assignment is handled by a
    GroupAssignent object.
    """

    def __init__(self, data_dict: Dict = None, style: str = "response"):
        """Return a completely built Policy object.

        The object can be created blank, or populated based on a Falcon API response dictionary.
        """
        # Configure the general settings that every policy will contain
        self.cid: str = None
        self.created_by: str = None
        self.created_timestamp: datetime = None
        self.description: str = None
        self.enabled: bool = None
        self.groups: List[GroupAssignment] = None
        self.modified_by: str = None
        self.modified_timestamp: str = None
        self.name: str = None
        self.platform_name: str = None
        self.policy_id: str = None
        self.settings_groups: List[PolicySettingGroup] = None
        self.settings_key_name = "settings"

        # TODO: Replace with a proper wrapper around ioa_rule_groups, this is temporary so that this
        #       result is still reachable from the 'dump' method when style="prevention".
        self._raw_ioa_rule_groups: List[Dict[str, Any]] = None

        # Set lists up here to ensure we do not pass references around when multiple Policy classes
        # are instantiated
        self.groups = []
        self.settings_groups = []

        # Customise the object so that it stores data appropriately based on the type of
        # policy that it will represent.
        if style == "response":
            self.settings_key_name = "settings"
        elif style == "prevention":
            self.prevention_settings = self.settings_groups
            self.settings_key_name = "prevention_settings"
        else:
            raise ValueError(f"The style provided ({style}) is not valid")

        self.style = style

        if data_dict:
            self._load_data_dict(data_dict=data_dict)

    def _load_data_dict(self, data_dict: Dict = None):
        """Load a policy dictionary from Falcon and set up the object accordingly."""
        self.cid = data_dict.get("cid")
        self.created_by = data_dict.get("created_by")
        self.created_timestamp = data_dict.get("created_timestamp")
        self.description = data_dict.get("description")
        self.enabled = data_dict.get("enabled")
        self.policy_id = data_dict.get("id")
        self.modified_by = data_dict.get("modified_by")
        self.modified_timestamp = data_dict.get("modified_timestamp")
        self.name = data_dict.get("name")
        self.platform_name = data_dict.get("platform_name")
        self._raw_ioa_rule_groups = data_dict.get("ioa_rule_groups")  # TODO: see earlier TODO

        # Load all groups as GroupAssignment objects
        groups: List[Dict] = data_dict.get("groups", [])
        for group_dict in groups:
            self.groups.append(GroupAssignment(data_dict=group_dict))

        # Load all groups of settings
        settings_groups: List[Dict] = data_dict.get(self.settings_key_name)
        for settings_group_dict in settings_groups:
            self.settings_groups.append(PolicySettingGroup(data_dict=settings_group_dict))

    def dump(self) -> Dict:
        """Return a dictionary representing a full policy.

        This dictionary is compliant with the CrowdStrike API specification, and therefore
        can replicate the exact response sent back by the CrowdStrike API if this policy
        already happened to exist in the Cloud.
        """
        dumped = {
            "cid": self.cid,
            "created_by": self.created_by,
            "created_timestamp": self.created_timestamp,
            "description": self.description,
            "enabled": self.enabled,
            "groups": [x.dump() for x in self.groups],
            "id": self.policy_id,
            "modified_by": self.modified_by,
            "modified_timestamp": self.modified_timestamp,
            "name": self.name,
            "platform_name": self.platform_name,
            self.settings_key_name: [x.dump() for x in self.settings_groups],
        }

        if self.style == "prevention":
            dumped["ioa_rule_groups"] = self._raw_ioa_rule_groups  # TODO: see earlier TODO

        return dumped

    def flat_dump(self) -> Dict:
        """Return a stripped down dictionary representing a full policy.

        This dictionary is designed for use with CrowdStrike API policy verbs (i.e., policy
        creation (POST) and modification (PATCH), and therefore is limited only to the content
        required to execute these actions.
        """
        settings: List[Dict] = []
        for settings_group in self.settings_groups:
            settings.extend(settings_group.flat_dump()['settings'])

        data = {
            "description": self.description,
            "name": self.name,
            "platform_name": self.platform_name,
            self.settings_key_name: settings,
        }
        if self.policy_id:
            data['id'] = self.policy_id

        return data



"""Caracara Hosts Module: host data history functions.

In order to avoid the main hosts.py file getting too unwieldly, this file contains
the implementations of the functions required to retrieve data that Falcon gathers
on hosts' network and login histories.
"""
from __future__ import annotations
from typing import (
    Dict,
    TYPE_CHECKING,
)

from caracara.common.batching import batch_get_data
from caracara.filters import FalconFilter
from caracara.filters.decorators import filter_string

if TYPE_CHECKING:
    from caracara.modules.hosts import HostsApiModule


@filter_string
def describe_network_address_history(
    self: HostsApiModule,
    filters: FalconFilter or str = None,
) -> Dict:
    """Return a dictionary of network address history for all devices in your Falcon tenant.

    Arguments
    ---------
    filters: FalconFilter or str, optional
        Filters to apply to the device search.

    Returns
    -------
    dict: A dictionary containing network address history for every device discovered.
    """
    device_ids = self.get_device_ids(filters)
    device_data = batch_get_data(device_ids, self.hosts_api.query_network_address_history)

    return device_data


@filter_string
def describe_login_history(
    self: HostsApiModule,
    filters: FalconFilter or str = None,
) -> Dict:
    """Return a dictionary containing login history for every device in your Falcon tenant.

    Arguments
    ---------
    filters: FalconFilter or str, optional
        Filters to apply to the device search.

    Returns
    -------
    dict: A dictionary containing login history for every device discovered.
    """
    self.logger.info("Describing login history for devices matching the filter: %s", filters)
    device_ids = self.get_device_ids(filters)
    device_data = batch_get_data(device_ids, self.hosts_api.query_device_login_history)

    return device_data


@filter_string
def describe_state(
    self: HostsApiModule,
    filters: FalconFilter or str = None,
) -> Dict[str, Dict]:
    """Return a dictionary containing online state for devices matching the provided filter.

    Arguments
    ---------
    filters: FalconFilter or str, optional
        Filters to apply to the device search.

    Returns
    -------
    dict: A dictionary containing state details for every device discovered.
    """
    self.logger.info("Describing device states according to the filter string %s", filters)
    device_ids = self.get_device_ids(filters)
    device_data = batch_get_data(device_ids, self.hosts_api.get_device_details)

    for device_id, state in batch_get_data(device_ids, self.hosts_api.get_online_state).items():
        if device_id in device_data:
            device_data[device_id]["state"] = state.get("state", "unknown")

    return device_data



"""Falcon Response Policies API."""
from functools import partial
from typing import Dict, List

from falconpy import (
    OAuth2,
    ResponsePolicies,
)

from caracara.common.decorators import platform_name_check
from caracara.common.module import FalconApiModule
from caracara.common.pagination import all_pages_numbered_offset_parallel
from caracara.common.policy_wrapper import Policy
from caracara.common.sorting import SORT_ASC, SORTING_OPTIONS
from caracara.filters import FalconFilter
from caracara.filters.decorators import filter_string
from caracara.modules.response_policies.template import generate_response_template


class ResponsePoliciesApiModule(FalconApiModule):
    """
    Response Policies API module.

    This module provides the logic to interact with the Falcon Response Policies API.
    """

    name = "CrowdStrike Response Policies API Module"
    help = "Describe, create, delete and edit Falcon Response policies"

    def __init__(self, api_authentication: OAuth2):
        """Construct an instance of the ResponsePoliciesApiModule class."""
        super().__init__(api_authentication)
        self.logger.debug("Configuring the FalconPy Response Policies API")
        self.response_policies_api = ResponsePolicies(auth_object=self.api_authentication)

    @filter_string
    def describe_policies_raw(
        self, filters: str or FalconFilter = None, sort: str = SORT_ASC
    ) -> List[Dict]:
        """Return a list of dictionaries containing all response policies in the Falcon tenant."""
        if sort not in SORTING_OPTIONS:
            raise ValueError("Sort must be SORT_ASC or SORT_DESC")

        self.logger.info("Describing all Falcon Response policies")
        partial_func = partial(
            self.response_policies_api.query_combined_policies,
            filter=filters,
            sort=sort,
        )
        resources = all_pages_numbered_offset_parallel(func=partial_func, logger=self.logger)
        self.logger.debug(resources)
        return resources

    @filter_string
    def describe_policies(
        self, filters: str or FalconFilter = None, sort: str = SORT_ASC
    ) -> List[Policy]:
        """Return a list of all response policies packaged as custom Python Policy objects."""
        raw_policies_dict = self.describe_policies_raw(filters=filters, sort=sort)
        policies: List[Policy] = []
        for policy_dict in raw_policies_dict:
            policy = Policy(data_dict=policy_dict, style="response")
            policies.append(policy)

        return policies

    @platform_name_check
    def new_policy(self, platform_name: str) -> Policy:  # pylint: disable=R0201
        """
        Return a platform-specific response policy template ready for customisation.

        To create the policy in the CrowdStrike cloud, use the push_policy function.
        """
        # We disable pylint rule R0201 as the decorator existence means that this function
        # cannot be a @staticmethod. The decorator expects the self parameter.
        return generate_response_template(platform_name=platform_name)

    def push_policy(self, policy: Policy) -> Policy:
        """Push a policy to the CrowdStrike Cloud, and return the resultant policy."""
        self.logger.info("Creating the response policy named %s", policy.name)
        response = self.response_policies_api.create_policies(body={
            "resources": [
                policy.flat_dump()
            ],
        })['body']
        new_policy = Policy(data_dict=response['resources'][0], style="response")
        return new_policy

    def add_policy_to_group(self, policy: Policy or str, group: str) -> Policy:
        """
        Add a policy to a group.

        Takes a policy object or policy ID and a group object or group ID as parameters.
        """
        if isinstance(policy, Policy):
            policy_id = policy.policy_id
        else:
            policy_id = policy

        # Eventually we will likely have support for host group objects and handle them here
        group_id = group

        action_parameters = [
            {
                "name": "group_id",
                "value": group_id,
            }
        ]

        self.response_policies_api.perform_policies_action(
            action_name="add-host-group",
            action_parameters=action_parameters,
            ids=[policy_id],
        )

        updated_policy_dict = self.response_policies_api.query_combined_policies(
            filter=f"id: '{policy_id}'",
        )['body']['resources'][0]

        new_policy = Policy(data_dict=updated_policy_dict, style="response")
        return new_policy

    def modify_policy(self, policy: Policy) -> Policy:
        """
        Modify a pre-existent policy in the CrowdStrike Cloud.

        Accepts a Policy object and patches the policy in the CrowdStrike Cloud to match
        the configuration of the passed Policy object.
        """
        if policy.policy_id is None:
            raise Exception(
                "You must supply a policy containing an ID. To retrieve a complete policy,"
                "use the describe_policies function"
            )

        self.logger.info("Updating the response policy named %s", policy.name)
        response = self.response_policies_api.update_policies(body={
            "resources": [
                policy.flat_dump()
            ],
        })['body']['resources'][0]
        new_policy = Policy(data_dict=response, style="response")
        return new_policy



"""Falcon Prevention Policies API."""
from functools import partial
from typing import List, Dict

from falconpy import (
    OAuth2,
    PreventionPolicies,
)

from caracara.common.decorators import platform_name_check
from caracara.common.module import FalconApiModule
from caracara.common.pagination import all_pages_numbered_offset_parallel
from caracara.common.policy_wrapper import Policy
from caracara.common.sorting import SORT_ASC, SORTING_OPTIONS
from caracara.filters import FalconFilter
from caracara.filters.decorators import filter_string
from caracara.modules.prevention_policies.template import generate_prevention_template


class PreventionPoliciesApiModule(FalconApiModule):
    """
    Prevention Policies API module.

    This module provides the logic to interact with the Falcon Prevention Policies API.
    """

    name = "CrowdStrike Prevention Policies API Module"
    help = "Describe, create, delete and edit Falcon Prevention policies"

    def __init__(self, api_authentication: OAuth2):
        """Construct an instance of the PreventionPoliciesApiModule class."""
        super().__init__(api_authentication)
        self.logger.debug("Configuring the FalconPy Prevention Policies API")
        self.prevention_policies_api = PreventionPolicies(auth_object=self.api_authentication)

    @filter_string
    def describe_policies_raw(
        self, filters: str or FalconFilter = None, sort: str = SORT_ASC
    ) -> List[Dict]:
        """Return a list of dictionaries containing all prevention policies in the Falcon tenant."""
        if sort not in SORTING_OPTIONS:
            raise ValueError("Sort must be SORT_ASC or SORT_DESC")

        self.logger.info("Describing all Falcon prevention policies")
        partial_func = partial(
            self.prevention_policies_api.query_combined_policies,
            filter=filters,
            sort=sort,
        )
        resources = all_pages_numbered_offset_parallel(func=partial_func, logger=self.logger)
        self.logger.debug(resources)
        return resources

    @filter_string
    def describe_policies(
        self, filters: str or FalconFilter = None, sort: str = SORT_ASC
    ) -> List[Policy]:
        """Return a list of all prevention policies packaged as custom Python Policy objects."""
        raw_policies_dict = self.describe_policies_raw(filters=filters, sort=sort)
        policies: List[Policy] = []
        for policy_dict in raw_policies_dict:
            policy = Policy(data_dict=policy_dict, style="prevention")
            policies.append(policy)

        return policies

    @platform_name_check
    def new_policy(self, platform_name: str) -> Policy:
        """
        Return a platform-specific prevention policy template ready for customisation.

        To create the policy in the CrowdStrike cloud, use the push_policy function.
        """
        return generate_prevention_template(platform_name=platform_name)

    def push_policy(self, policy: Policy) -> Policy:
        """Push a policy to the CrowdStrike Cloud, and return the resultant policy."""
        self.logger.info("Creating the prevention policy named %s", policy.name)
        response = self.prevention_policies_api.create_policies(body={
            "resources": [
                policy.flat_dump()
            ],
        })['body']
        new_policy = Policy(data_dict=response['resources'][0], style="prevention")
        return new_policy



"""
Falcon FQL Filter Wrapper class.

This class keeps track of Falcon FQL filter objects, and can return a valid, fully generated
FQL string representing all the filter objects within it.
"""
import logging

from typing import (
    Any,
    Dict,
    List,
)
from uuid import uuid4

from caracara.filters.fql import FalconFilterAttribute


class FalconFilter:
    """
    Falcon FQL Filter Wrapper class.

    Class to handle filtering hosts within Falcon. This encompasses the various known types of
    filter, and allows building up FQL queries, ready to pass straight into Falcon API calls
    as GET parameters.

    Each filter is referred to within the class by a unique, randomly generated ID. This is beacuse
    it is possible to combine filters of the same type (e.g., to have a date range), so the filter
    type cannot be used as a key within the internal dictionary.
    """

    available_filters: Dict[str, FalconFilterAttribute] = {}
    logger = logging.getLogger(__name__)

    def set_filter_value(self, filter_id: str, value):
        """Set the value for a filter by filter ID."""
        self.logger.info("Setting filter %s to %s", filter_id, value)
        if filter_id not in self.filters:
            raise Exception(f"Filter with ID {filter_id} does not exist here")

        self.filters[filter_id].set_value(value)

    def set_filter_operator(self, filter_id: str, operator: str):
        """Set the operator / comparator for an FQL filter by filter ID."""
        self.logger.info("Setting the filter operator of filter %s to %s", filter_id, operator)
        if filter_id not in self.filters:
            raise Exception(f"Filter with ID {filter_id} does not exist here")

        self.filters[filter_id].set_operator(operator)

    def add_filter(self, new_filter: FalconFilterAttribute) -> str:
        """Add a filter built externally to this object to the internal filter dictionary."""
        filter_id = str(uuid4())
        self.logger.info("Adding a new %s filter with filter ID %s", new_filter.name, filter_id)
        self.filters[filter_id] = new_filter
        return filter_id

    def remove_filter(self, filter_id: str):
        """Remove a filter from the current filter object by filter ID."""
        if filter_id in self.filters:
            self.logger.info("Removing filter with ID %s", filter_id)
            del self.filters[filter_id]
        else:
            self.logger.info(
                "Attempted to remove filter with ID %s, but it does not exist",
                filter_id,
            )

    def create_new_filter(
        self,
        filter_name: str,
        initial_value: Any = None,
        initial_operator: str = "EQUAL",
    ) -> str:
        """
        Create a new filter and add it to the FalconFilter object.

        This function will create a new filter via the filter's friendly name.
        The new filter's unique ID string is returned in case the developer wishes
        to reference and/or remove this specific filter within the FalconFilter object later.
        """
        self.logger.debug(
            "Creating a new %s filter (initial operator: %s; initial value: %s",
            filter_name, initial_operator, initial_value,
        )
        if filter_name not in self.available_filters:
            raise Exception(filter_name)

        new_filter: FalconFilterAttribute = self.available_filters[filter_name]()
        if initial_value is not None:
            new_filter.set_value(initial_value)
        if initial_operator:
            new_filter.set_operator(initial_operator)

        filter_id = self.add_filter(new_filter)
        self.logger.info("Created new %s filter with ID %s", filter_name, filter_id)
        return filter_id

    def create_new_filter_from_kv_string(self, key_string: str, value) -> str:
        """
        Create a filter from a key->value string.

        Examples:
        -> Domain__NOT=ExcludeDomain.com
        -> LastSeen__GTE=1970-01-01T00:00:00Z
        """
        self.logger.info("Loading filter from KV string: %s", key_string)

        if "__" in key_string:
            filter_name, operator = key_string.split("__")
        else:
            filter_name = key_string
            operator = None  # None operator results in the default

        if isinstance(value, str):
            if ',' in value:
                value = value.split(',')

        if operator:
            return self.create_new_filter(
                filter_name=filter_name,
                initial_value=value,
                initial_operator=operator
            )

        return self.create_new_filter(
            filter_name=filter_name,
            initial_value=value
        )

    def get_fql(self) -> str:
        """Return a valid FQL string based on the filters within the FalconFilter object."""
        fql = '+'.join(x.get_fql() for x in self.filters.values())
        self.logger.info("Generated FQL: %s", fql)
        return fql

    def __init__(self, filters: List[FalconFilterAttribute] = None):
        """
        Set up a new Falcon Filter object.

        If filter data already exists, this can be passed in as a dictionary so that the
        FQL can be generated instantly. Otherwise, the developer can add filter data at will,
        and then dump to FQL when ready.
        """
        self.filters: Dict[str, FalconFilterAttribute] = {}

        self.logger.debug("Initialising a new FalconFilter object")
        if filters:
            for initial_filter in filters:
                self.add_filter(initial_filter)

    def __str__(self):
        """Dump the filters within the FalconFilter object to an FQL string."""
        return self.get_fql()



r"""Caracara: Pagination functions.

MM'''''''`YM                   oo                     dP
MM  mmmmm  M                                          88
M'        .M .d8888b. .d8888b. dP 88d888b. .d8888b. d8888P .d8888b. 88d888b.
MM  MMMMMMMM 88'  `88 88'  `88 88 88'  `88 88'  `88   88   88'  `88 88'  `88
MM  MMMMMMMM 88.  .88 88.  .88 88 88    88 88.  .88   88   88.  .88 88
MM  MMMMMMMM `88888P8 `8888P88 dP dP    dP `88888P8   dP   `88888P' dP
MMMMMMMMMMMM               .88
                       d8888P                       for CrowdStrike Caracara

This file contains code for all types of pagination styles used by the
Falcon APIs, to avoid code repetition within the individual modules.

The endpoints handled here typically return a list of IDs based on some
kind of filter operation. If you are looking for parallised batching
of data dictionaries returned in response to a list of IDs, the batching
code file contains the required code to pull this data down as quickly
as possible.

Four types of paginator are in use:

Style 1 (Numerical Offset)
Implementation function: all_pages_numbered_offset()

We paginate by providing a limit (default: 100), and grabbing pages with
that number of items per page. Keep going until len(list of items) == total

This is optimised for performance by grabbing multiple pages in parallel.

Pages are structured like this
{
    "meta": {
        "pagination": {
            "limit": items_per_page_maximum: int,
            "offset": items_to_skip: int,
            "total": total_number_of_items: int,
        }
    }
}


Style 2 (Token Offset)
Implementation function: all_pages_token_offset()

In this style, a token is returned by the server. When we send the token
to the server, we get the next page of content as well as another token.
We keep requesting data until we have all the items.

It is not possible to parallise these requests because each page depends
on the retrieval of its page token from the previous page.

Pages are structured like this
{
    "meta": {
            "limit": items_per_page_maximum: int,
            "offset": page_token: str,
            "total": total_number_of_items: int,
    }
}


Style 3 (Token After)
This style works exactly the same as Style 2, except the token is stored in
an attribute named `after`.
Implementation function: all_pages_token_offset(offset_key_named_after=True)


Style 4 (Token _marker)
Some APIs (such as those associated with our Intel service) divide data up by
times, rather than by pages of data. In these cases, an FQL filter named
_marker is used to indicate the position within the result set.
Once a Caracara wrapper to such an API is implemented, a function to handle
this will be built.
See the "Deep Pagination Leveraging Markers (Timestamp)" section here:
    https://falconpy.io/Usage/Response-Handling.html
"""
import concurrent.futures
import logging

from functools import partial
from threading import current_thread
from typing import Callable, Dict, List

from caracara.common.batching import batch_data_pull_threads
from caracara.common.constants import PAGINATION_LIMIT, SCROLL_BATCH_SIZE


def all_pages_numbered_offset(
    func: Callable[[Dict[str, Dict]], List[Dict] or List[str]],
    logger: logging.Logger,
    body: Dict = None,
    limit: int = PAGINATION_LIMIT,
):
    """Grab all pages from a paginated call in Style 1 (numerical offset)."""
    logger = logger.getChild("pagination_numbered_offset")
    logger.info(f"Grabbing all pages from the {func.__name__} function (limit: {limit})")

    offset = 0
    found_all = False
    all_resources = []

    while not found_all:
        logger.info(f"Grabbing batch of items {offset + 1} up to {offset + limit}")
        req_body = {
            "limit": limit,
            "offset": offset,
        }
        if body:
            req_body.update(body)

        response = func(body=req_body)['body']
        resources = response['resources']
        logger.info(f"Retrieved a batch of {len(resources)} items")
        logger.debug(response)
        all_resources.extend(resources)

        # If there are no resources returned, we may get a KeyError later
        if not all_resources:
            return []

        if response['meta']['pagination']['total'] > len(all_resources):
            offset = response['meta']['pagination']['offset']
        else:
            found_all = True

    return all_resources


def _numbered_offset_parallel_worker(
    func: Callable[[Dict[str, Dict]], List[Dict] or List[str]] or partial,
    limit: int,
    logger: logging.Logger,
    batch_offset: int,
) -> List[Dict]:
    thread_name = current_thread().name
    if isinstance(func, partial):
        logger.info(
            "%s | Batch worker started with limit %d and offset %d; partial function: %s",
            thread_name, limit, batch_offset, func.func.__name__,
        )
    else:
        logger.info(
            "%s | Batch worker started with offset %s; function: %s",
            thread_name, limit, func.__name__,
        )
    response = func(offset=batch_offset, limit=limit)['body']
    logger.debug("%s | %s", thread_name, response)
    resources = response['resources']
    logger.info("%s | Retrieved %d resources", thread_name, len(resources))

    return resources


def all_pages_numbered_offset_parallel(
    func: Callable[[Dict[str, Dict]], List[Dict] or List[str]] or partial,
    logger: logging.Logger,
    limit: int = PAGINATION_LIMIT,
) -> List[Dict]:
    """Grab all pages from a paginated call in Style 1 (numerical offset), super charged."""
    logger = logger.getChild(__name__)
    if isinstance(func, partial):
        logger.info(
            "Pagination Style 1: Grabbing all pages from the partial %s function (limit: %d)",
            func.func.__name__, limit,
        )
    else:
        logger.info(
            "Pagination Style 1: Grabbing all pages from the %s function (limit: %d)",
            func.__name__, limit,
        )

    # Get the first page to figure out how many items to pull
    logger.info("Grabbing first batch of items 1 to up to %s", limit)
    response = func(offset=0, limit=limit)['body']
    logger.debug(response)

    resources = response['resources']
    logger.info("Retrieved a batch of %d items", len(resources))
    all_resources: List[Dict] = resources

    if not all_resources:
        logger.info("No resources received; returning an empty list")
        return []

    total = response['meta']['pagination']['total']
    logger.info("Total number of resources: %s", total)
    if total == len(all_resources):
        # Received all resources in the first batch
        return all_resources

    """
    List of limits to request from Falcon (e.g., for 301 items, with
    a limit of 100):
    batches = [100,        200,        300]
                ^           ^           ^
               pg2         pg3         pg4
        items 100 - 199  200 - 299  300 - 301

    We already have pg1 (0 - 99) from the first API call
    """
    batches = list(range(limit, total, limit))
    logger.info("Divided offsets into %d of %d", len(batches), limit)
    logger.debug(batches)

    threads = batch_data_pull_threads()

    with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
        partial_worker = partial(_numbered_offset_parallel_worker, func, limit, logger)
        completed = executor.map(
            partial_worker,
            batches,
        )

    for complete in completed:
        logger.debug("Completed a batch")
        all_resources.extend(complete)

    logger.debug("Returning resource list")
    logger.debug(all_resources)
    return all_resources


def all_pages_token_offset(
    func: Callable[[Dict[str, Dict]], List[Dict] or List[str]] or partial,
    logger: logging.Logger,
    limit: int = SCROLL_BATCH_SIZE,
    offset_key_named_after: bool = False,
) -> List:
    """Grab all pages from a token offset-based pagination endpoint.

    This defaults to implenting Style 2, which is a scroll-style paginatied endpoint
    that takes a token as an offset. However, if you set offset_key_named_after to True,
    this endpoint implements pagination Style 3.
    """
    logger = logger.getChild(__name__)
    if isinstance(func, partial):
        logger.info(
            "Pagination Style 2: Grabbing all pages from the partial %s function (limit: %d)",
            func.func.__name__, limit,
        )
    else:
        logger.info(
            "Pagination Style 2: Grabbing all pages from the %s function (limit: %d)",
            func.__name__, limit,
        )

    complete = False
    item_ids = []
    offset = None

    current_page = 0

    while not complete:
        current_page += 1
        logger.info(
            "Fetching page %d: %d to up to %d",
            current_page, len(item_ids) + 1, limit * current_page,
        )
        if offset_key_named_after:
            response = func(limit=limit, after=offset)['body']
        else:
            response = func(limit=limit, offset=offset)['body']
        logger.debug(response)
        item_ids.extend(response['resources'])
        if not item_ids:
            # Nothing was returned, so bail out in case the pagination data does not exist
            return []

        pagination_data = response['meta']['pagination']
        if pagination_data['total'] > len(item_ids):
            if offset_key_named_after:
                offset = pagination_data['after']
            else:
                offset = pagination_data['offset']
        else:
            complete = True

    logger.debug("Returned IDs:")
    logger.debug(item_ids)

    return item_ids



# pylint: disable=too-few-public-methods
"""Common functions available to filters across all modules."""
import re

from datetime import datetime, timedelta


class IPChecker:
    """Check IP address strings for validity via regex."""

    def __init__(self):
        """Set up an IP address checked object and pre-compile the regex."""
        self._ipv4_re = re.compile(r"^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$")

    def check_ipv4(self, addr: str) -> bool:
        """Check whether an IPv4 address matches the specification regex."""
        return self._ipv4_re.match(addr) is not None


class ISOTimestampChecker:
    """Check timestamp strings to see whether they conform to the ISO8601 UTC Zulu format."""

    def __init__(self):
        """Set up an ISO timestamp checker object and pre-compile the regex."""
        self._iso_re = re.compile(
            r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z'
        )

    def is_iso_format(self, timestamp: str):
        """Perform the check based on the regex expression compiled within the class."""
        return self._iso_re.match(timestamp) is not None


class RelativeTimestamp:
    """
    Relative Timestamp generator class.

    This class contains the logic required to convert from a string (e.g., -30m) to an absolute
    UTC ISO8601 timestamp. This is the format expected by the Falcon API, and enables filters
    such as LastSeen.

    Examples:
    -1hr = take one hour away from the current time
    -30m = take thirty mins away from the current time
    -2d  = take 2 days away from the current time
    +4d  = add 4 days to the current time
    """

    def __init__(self):
        """Set up a RelativeTimestamp object and pre-compile the regex."""
        self._relative_re = re.compile(
            r"^(?P<sign>[-+])(?P<number>\d+)(?P<scale>(s|m|h|d))$"
        )

    def check_relative_timestamp(self, relative_timestamp: str) -> bool:
        """Check whether a relative timestamp matches the specification regex."""
        return self._relative_re.match(relative_timestamp) is not None

    def convert_relative_timestamp(
        self,
        original_timestamp: datetime,
        relative_timestamp: str
    ) -> datetime:
        """Convert a relative timestamp into an absolute ISO8601 timestamp."""
        match = self._relative_re.match(relative_timestamp)
        if match is None:
            # This should be impossible, as we have the check function
            # above to make sure this ridiculous situation doesn't happen
            raise Exception("The timestamp did not match the prescribed format")

        sign = match.group('sign')
        number = int(match.group('number'))
        scale = match.group('scale')

        # Convert to seconds to save code and effort
        if scale == 's':
            seconds = number
        elif scale == 'm':
            seconds = number * 60
        elif scale == 'h':
            seconds = number * 60 * 60
        elif scale == 'd':
            seconds = number * 60 * 60 * 24
        else:
            # Assuming the regex did The Thing, this should be impossible
            raise Exception("The relative timestamp did not contain a supported unit")

        if sign == '-':
            new_timestamp = original_timestamp - timedelta(seconds=seconds)
        else:
            new_timestamp = original_timestamp + timedelta(seconds=seconds)

        return new_timestamp



"""Filter-Related Function Decorators."""
from functools import wraps
from inspect import signature
from typing import Callable

from caracara.filters import FalconFilter


def filter_string(func: Callable):
    """Decorate a function to ensure that a Falcon Filter object is converted to an FQL string."""
    # Load the function's signature, and confirm that self and filters are present
    sig = signature(func)
    if 'filters' not in sig.parameters:
        raise ValueError(f'The function {func.__name__} does not have a filters parameter')

    if 'self' not in sig.parameters:
        raise ValueError(
            f"The function {func.__name__} must be a class function with a self parameter"
        )

    @wraps(func)
    def wrapper(*args, **kwargs):
        # Load in the arguments passed to the function and bind them to the function's signature
        _args = sig.bind(*args, **kwargs)
        # Apply any default parameters (e.g., the filters=None where filters is not specified)
        _args.apply_defaults()
        # Get references to the self and filters arguments for logging and processing purposes
        self = _args.arguments['self']
        filters = _args.arguments['filters']

        if hasattr(self, 'logger'):
            self.logger.debug(f"Entering filter_string wrapper for function {func.__name__}")

        if not filters:
            filter_str = None
        elif isinstance(filters, FalconFilter):
            if hasattr(self, 'logger'):
                self.logger.debug("Rewriting FalconFilter to a string via its get_fql() function")
            filter_str = filters.get_fql()
        elif isinstance(filters, str):
            filter_str = filters
        else:
            raise Exception(
                "A filter of an incorrect type was passed into the function. "
                "Expected a FalconFilter or a string"
            )

        if hasattr(self, 'logger'):
            self.logger.debug("Exiting decorator")

        # Overwrite the filters argument in the bound signature with the string representation.
        # This ensures that FalconPy is always passed a string, and never a FalconFilter object.
        _args.arguments['filters'] = filter_str

        # Return the original function with the modified arguments
        return func(*_args.args, **_args.kwargs)
    return wrapper



"""Falcon Query Language (FQL) generation code."""
from abc import ABC, abstractmethod
from typing import List

from caracara.common.constants import FILTER_OPERATORS


class FalconFilterAttribute(ABC):
    """
    Generic Falcon Filter Attribute.

    Class to represent an something that can be filtered on, such as a
    computer domain, OS or computer role (e.g. Server, DC, Workstation).
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """Store the friendly name for the filter. Cannot include spaces. Example: OS."""

    @property
    def description(self) -> str:
        """
        Store a short description of the filter and what it can be used for.

        Defaults to returning the doc string.
        """
        return __doc__

    @property
    @abstractmethod
    def fql(self) -> str:
        """Return the name of the attribute as it will be written when converted to FQL."""

    def example(self) -> str:
        """
        Store example usage of the filter.

        Auto-generated based on options, but can be overridden with a simple string function.
        """
        if not hasattr(self, 'options'):
            raise NotImplementedError(
                "Filter should implement its example string as no set "
                "options are programmed into this filter."
            )

        if isinstance(self.options, list):
            return (
                "This filter accepts one or more of the following "
                "options as a comma delimited list or bare string: " +
                ", ".join(self.options)
            )

        if isinstance(self.options, dict):
            return (
                "This filter accepts one or more of the following "
                "options as a comma delimited list or bare string: " +
                ", ".join(self.options.keys())
            )

        # We have reached a self.options type that we do not support yet (or is incorrect)
        raise NotImplementedError(
            "Filter should implement its example string as the options "
            "are of an unknown type."
        )

    # Define the acceptable types for this filter's value passed in by the
    # code or user
    types: List = [str, list]

    # Default to not restricting the values to a predefined list
    restrict: bool = False

    # A list of valid values
    options: None

    # The current value of this filter instance
    value = None

    # Valid operators (e.g. EQUAL, NOT, etc.)
    valid_operators: List[str] = [
        "EQUAL",
        "NOT",
    ]

    # Current operator, which defaults to EQUAL unless overridden
    operator = "EQUAL"

    def __str__(self):
        """Convert a filter attribute to a valid FQL string."""
        return f'{self.name}: {self.operator} {self.value}'

    def _options_is_dict(self) -> bool:
        if isinstance(self.options, dict):
            return True

        if isinstance(self.options, list):
            return False

        # self.options is of an invalid type according to the specification
        raise Exception(
            f'The filter {self.name} was initialised incorrectly'
        )

    def _check_option(self, value: str) -> bool:
        if self.options is None:
            raise Exception(
                (
                    f'The filter {self.name} was initialised '
                    'incorrectly. options cannot be None here.'
                )
            )

        if self._options_is_dict():
            if value in self.options.keys():
                return True
        else:
            if value in self.options:
                return True

        return False

    def _check_value(self, value) -> bool:
        """Check whether a certain value will fulfill the criteria given."""
        # If the type of the value is invalid then the value is automatically
        # marked as invalid
        if not any((isinstance(value, x) for x in self.types)):
            return False

        # If the type of the value is valid and the value itself is not
        # restricted to a fixed list, then the value is okay
        if not self.restrict:
            return True

        # We now cover the main cases (string or list of strings);
        # therefore, if the types list above is changed for a given subclass
        # then this function will need to be overridden
        if isinstance(value, str):
            return self._check_option(value)

        if isinstance(value, list):
            for sub_value in value:
                if not isinstance(sub_value, str):
                    return False
                if not self._check_option(sub_value):
                    return False
            return True

        # The filter value is not an instance of a supported type
        raise Exception(
            f'The filter {self.name} was initialised incorrectly. '
            '_check_value must be overridden to use a different type of '
            'value coercion.'
        )

    def set_value(self, value):
        """Set the filter value utilising internal sanity checks."""
        if self._check_value(value):
            self.value = value
        else:
            raise Exception(
                self.name,
                str(value)
            )

    def set_operator(self, operator: str):
        """Set the operator to be applied to the filter (e.g., =, >)."""
        if operator in self.valid_operators:
            self.operator = operator
        else:
            raise Exception(
                self.name,
                operator,
                self.valid_operators
            )

    def get_fql(self) -> str:
        """
        Convert the internal data structure to useful FQL.

        For different types of filters this function may need to be overridden.
        It will work for str, list(str) and bool filters out the box.
        """
        operator_symbol = FILTER_OPERATORS[self.operator]

        if self.value is None:
            return f'{self.fql}:{operator_symbol}null'

        if self.restrict and self._options_is_dict():
            if isinstance(self.value, str):
                fql_value = "'" + self.options[self.value] + "'"
            else:
                fql_value = (
                    '[' +
                    ','.join(
                        f"'{self.options[x]}'" for x in self.value
                    ) +
                    ']'
                )
        else:
            if isinstance(self.value, str):
                fql_value = f"'{self.value}'"
            elif isinstance(self.value, bool):
                fql_value = f"{str(self.value).lower()}"
            else:
                fql_value = (
                    '[' +
                    ','.join(
                        f"'{x}'" for x in self.value
                    ) +
                    ']'
                )
        fql_string = f'{self.fql}: {operator_symbol}{fql_value}'
        return fql_string



"""Caracara batching support.

This module contains all the required code to generically parallelise retrieval of data batches.

The endpoints handled here typically look for a list of IDs in the body
and return back a dictionary of data.

For paginated results that send back IDs, the pagination code file contains
the required code to pull this data down as quickly as possible.
"""
import concurrent.futures
import logging
import multiprocessing

from functools import partial
from threading import current_thread
from typing import Callable, Dict, List

from caracara.common.constants import DATA_BATCH_SIZE


BATCH_LOGGER = logging.getLogger(__name__)


def batch_data_pull_threads() -> int:
    """
    Return maximum number of threads to spin up in a thread pool.

    Result is either 20, or 2x the number of cores, whichever is smallest.
    """
    cores = multiprocessing.cpu_count()
    thread_count = min(cores * 2, 20)
    BATCH_LOGGER.debug("Will use up to %d threads for batch data retrieval", thread_count)
    return thread_count


def batch_get_data(
    lookup_ids: str,
    func: Callable[[object, List[str], str], Dict],
) -> Dict[str, Dict]:
    """Retrieve details for the list of AIDs provided.

    Arguments
    ---------
    lookup_ids: str
        List of IDs to retrieve data for.
    func: Python function, object
        Method to call by each created thread.

    Returns
    -------
    dict: A dictionary of all results returned.
    """
    resources = []

    BATCH_LOGGER.info("Batch data retrieval for %s (%d items)", func.__name__, len(lookup_ids))
    BATCH_LOGGER.debug(str(lookup_ids))

    # Divide the list of item IDs into a list of lists, each of size DATA_BATCH_SIZE
    batches = [lookup_ids[i:i+DATA_BATCH_SIZE] for i in range(0, len(lookup_ids), DATA_BATCH_SIZE)]
    BATCH_LOGGER.info("Divided the item IDs into %d batches", len(batches))

    threads = batch_data_pull_threads()

    def worker(batch_func: Callable[[List[str]], Dict], worker_lookup_ids: List[str]) -> Dict:
        thread_name = current_thread().name
        BATCH_LOGGER.info(
            "%s | Batch worker started with a list of %d items. Function: %s",
            thread_name, len(worker_lookup_ids), batch_func.__name__,
        )
        resources = batch_func(ids=worker_lookup_ids)["body"]["resources"]
        BATCH_LOGGER.info("%s | Retrieved %d resources", thread_name, len(resources))
        BATCH_LOGGER.debug("%s | %s", thread_name, resources)

        return resources

    with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
        partial_worker = partial(worker, func)
        completed = executor.map(partial_worker, batches)

    for complete in completed:
        BATCH_LOGGER.debug("Completed a batch")
        resources.extend(complete)

    resources_dict = {}
    for resource in resources:
        if 'id' in resource:
            resources_dict[resource['id']] = resource
        elif 'device_id' in resource:
            resources_dict[resource['device_id']] = resource
        elif 'child_cid' in resource:
            resources_dict[resource['child_cid']] = resource
        else:
            raise Exception("No ID field to build the dictionary from")

    BATCH_LOGGER.debug("Returned resources")
    BATCH_LOGGER.debug(resources_dict)

    return resources_dict



"""Caracara FQL filters that can be used in multiple modules."""
from caracara.filters.fql import FalconFilterAttribute


class GenericNameFilterAttribute(FalconFilterAttribute):
    """Generic filter for any 'name' attribute."""

    name = "Name"
    fql = "name"

    def example(self) -> str:
        """Show filter example."""
        return (
            "This filter accepts any string to be passed as a 'name' "
            "attribute. Examples of named objects in FQL include "
            "response and prevention policies."
        )


FILTER_ATTRIBUTES = [
    GenericNameFilterAttribute,
]



"""Caracara module-agnostic decorators."""
from functools import wraps
from inspect import signature
from typing import Callable

from caracara.common.constants import PLATFORMS


def platform_name_check(func: Callable):
    """Decorate a function to ensure that a platform_name argument is within the specified list."""
    # Load the function's signature, and confirm that the platform_name parameter has been added
    sig = signature(func)
    if 'platform_name' not in sig.parameters:
        raise ValueError(f'The function {func.__name__} does not have a platform_name parameter')

    if 'self' not in sig.parameters:
        raise ValueError(
            f"The function {func.__name__} must be a class function with a self parameter"
        )

    @wraps(func)
    def wrapper(*args, **kwargs):
        # Load in the arguments passed to the function and bind them to the function's signature
        _args = sig.bind(*args, **kwargs)
        # Apply any default parameters (e.g., platform_name=None where this is not specified)
        _args.apply_defaults()
        # Get references to the self and platform_name arguments for logging and processing purposes
        self = _args.arguments['self']
        platform_name = _args.arguments['platform_name']

        if hasattr(self, 'logger'):
            self.logger.debug(f"Entering filter_string wrapper for function {func.__name__}")

        if platform_name and isinstance(platform_name, str):
            if platform_name not in PLATFORMS:
                raise Exception(
                    f"The platform_name value supplied ({platform_name}) is not valid. "
                    f"Valid options are: {PLATFORMS}"
                )

        # If we get this far, either the platform_name given is reasonable, or it was None (which
        # is reasonable if the function signature allows this)

        if hasattr(self, 'logger'):
            self.logger.debug("Exiting decorator")

        # Return to the original function
        return func(*_args.args, **_args.kwargs)
    return wrapper



r"""Caracara base API interface client.

CrowdStrike Caracara
    _______ __ __             __
   |   _   |  |__.-----.-----|  |_
   |.  1___|  |  |  -__|     |   _|
   |.  |___|__|__|_____|__|__|____|
   |:  1   |
   |::.. . |     for FalconPy
   `-------'
"""

import logging

try:
    from falconpy import (
        OAuth2,
        confirm_base_region,
        confirm_base_url
    )
except ImportError as no_falconpy:
    raise SystemExit("The crowdstrike-falconpy library is not installed.") from no_falconpy

from caracara.common import FILTER_ATTRIBUTES as common_filter_attributes
from caracara.common.interpolation import VariableInterpolator
from caracara.common.meta import user_agent_string
from caracara.filters.falcon_filter import FalconFilter
from caracara.filters.fql import FalconFilterAttribute
from caracara.modules import (
    CustomIoaApiModule,
    FlightControlApiModule,
    HostsApiModule,
    PreventionPoliciesApiModule,
    ResponsePoliciesApiModule,
    RTRApiModule,
    MODULE_FILTER_ATTRIBUTES,
)


class Client:
    """Falcon API client.

    This class exposes all available Falcon API commands, and proxies requests through
    the FalconPy library.
    """

    class FalconFilter(FalconFilter):
        """Falcon Filter wrapper class.

        Create a sub-class of the FalconFilter class which we own locally within
        the client. This allows us to dynamically calculate the FQL filter attributes
        that we hava available and make them available at runtime.
        """

    def __init__(  # pylint: disable=R0913,R0914,R0915
        self,
        client_id: str = None,
        client_secret: str = None,
        cloud_name: str = "auto",
        member_cid: str = None,
        ssl_verify: bool = True,
        timeout: float = None,
        proxy: str = None,
        user_agent: str = None,
        verbose: bool = False,
        falconpy_authobject: OAuth2 = None,
    ):
        """Configure a Caracara Falcon API Client object."""
        self.logger = logging.getLogger(__name__)

        if client_id is None and client_secret is None and falconpy_authobject is None:
            raise Exception(
                "You must provide either a Client ID and Client Secret, "
                "or a pre-created FalconPy OAuth2 object"
            )

        if client_id is not None and client_secret is None:
            raise Exception("You cannot provide a Client ID without a Client Secret")

        if client_id is not None and falconpy_authobject is not None:
            raise Exception(
                "Please provide either a Client ID/Client Secret pair, "
                "or a pre-created FalconPy OAuth2 object, but not both"
            )

        self.logger.info("Setting up the Caracara client and configuring authentication")

        if client_id:
            # Load all parameters for the FalconPy authentication object into a dictionary
            # and handle environment variable interpolation
            interpolator = VariableInterpolator()
            cloud_name = interpolator.interpolate(cloud_name)
            client_id = interpolator.interpolate(client_id)
            client_secret = interpolator.interpolate(client_secret)
            member_cid = interpolator.interpolate(member_cid)
            user_agent = interpolator.interpolate(user_agent)
            proxy = interpolator.interpolate(proxy)
            user_agent = interpolator.interpolate(user_agent)

            if user_agent:
                user_agent = f"{user_agent} ({user_agent_string()})"
            else:
                user_agent = user_agent_string()

            self.logger.debug("User agent: %s", user_agent)

            auth_keys = {
                "base_url": cloud_name,
                "client_id": client_id,
                "client_secret": client_secret,
                "member_cid": member_cid,
                "proxy": proxy,
                "user_agent": user_agent,
                "ssl_verify": ssl_verify,
                "timeout": timeout,
            }

            self.logger.info(
                "Client ID: %s; Cloud: %s; Member CID: %s",
                client_id, cloud_name, member_cid,
            )
            self.logger.debug("SSL verification is %s", ssl_verify)
            self.logger.debug("Timeout: %s", str(timeout))
            self.logger.debug("Configured proxy: %s", proxy)

            # Remove all None values, as we do not wish to override any FalconPy defaults
            for k in list(auth_keys.keys()):
                if auth_keys[k] is None:
                    del auth_keys[k]

            self.verbose = verbose
            self.logger.debug("Verbose mode: %s", verbose)

            base_url = confirm_base_region(confirm_base_url(provided_base=cloud_name))
            self.logger.info("Base URL: %s", base_url)

            self.logger.debug("Configuring api_authentication object as an OAuth2 object")
            self.api_authentication = OAuth2(**auth_keys)
        elif falconpy_authobject:
            self.logger.info(
                "Using pre-created FalconPy OAuth2 object. All other options will be ignored"
            )
            self.api_authentication = falconpy_authobject
        else:
            raise Exception("Impossible authentication scenario")

        self.logger.info("Requesting API token")
        self.api_authentication.token()  # Need to force the authentication to resolve the base_url
        self.logger.info("Resolved Base URL: %s", self.api_authentication.base_url)

        # Configure modules here so that IDEs can pick them up
        self.logger.debug("Setting up Custom IOA module")
        self.custom_ioas = CustomIoaApiModule(self.api_authentication)
        self.logger.debug("Setting up the Flight Control module")
        self.flight_control = FlightControlApiModule(self.api_authentication)
        self.logger.debug("Setting up the Hosts module")
        self.hosts = HostsApiModule(self.api_authentication)
        self.logger.debug("Setting up the Prevention Policies module")
        self.prevention_policies = PreventionPoliciesApiModule(self.api_authentication)
        self.logger.debug("Setting up the Response Policies module")
        self.response_policies = ResponsePoliciesApiModule(self.api_authentication)
        self.logger.debug("Setting up the RTR module")
        self.rtr = RTRApiModule(self.api_authentication)

        self.logger.debug("Configuring FQL filters")
        # Pre-configure the FQL modules for faster instantiation later
        filter_attribute_classes = [
            *common_filter_attributes,
            *MODULE_FILTER_ATTRIBUTES,
        ]
        available_filters = {}
        for filter_attribute_class in filter_attribute_classes:
            instance: FalconFilterAttribute = filter_attribute_class()
            available_filters[instance.name] = filter_attribute_class

        # Modify the falcon filter class to contain the available filters
        self.FalconFilter.available_filters = available_filters

        self.logger.info("Caracara client configured")

    def __enter__(self):
        """Allow for entry as a context manager."""
        self.logger.debug("Entering Caracara context manager")
        return self

    def __exit__(self, *args):
        """Discard our token when we exit the context."""
        self.logger.info("Discarding API token on Client exit")
        self.api_authentication.revoke(self.api_authentication.token_value)
        self.logger.debug("Discarding Caracara context manager")
        return args



"""Generic Caracara API module.

This module contains the the FalconApiModule class, which represents a generic
Caracara API module. All modules, including Hosts, Prevention Policies, etc.
derive from this abstract base class.
"""
import logging

from abc import ABC, abstractmethod

from falconpy import OAuth2


class FalconApiModule(ABC):
    """
    Meta class for a generic Caracara API Module.

    Each module provides API Methods.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """Store the name for the developer to use when instantiating the API module."""

    @property
    @abstractmethod
    def help(self) -> str:
        """Store the help string to be made available for each API module."""

    def __init__(self, api_authentication: OAuth2):
        """Configure a Caracara API module with a FalconPy OAuth2 module."""
        class_name = self.__class__.__name__
        self.logger = logging.getLogger(class_name)
        self.logger.debug("Initialising API module: %s", class_name)

        self.api_authentication = api_authentication



r"""Caracara: Environment Variable Interpolator.

     _______                ___ ___ __          __   _____
    |   _   .---.-.--.--.  |   Y   |  |--.---.-|  |_|__   |
    |   1___|  _  |  |  |  |.  |   |     |  _  |   _',  ,-
    |____   |___._|___  |  |. / \  |__|__|___._|____||--|
    |:  1   |     |_____|  |:      |                 '--
    |::.. . |              |::.|:. |
    `-------'              `--- ---'

The functions within this file allow for interpolation of variables based on
the contents of environment variables, similar to how Docker allows for
strings such as ${SOME_VALUE} within Dockerfiles.

You can then do something along the lines of:
client = Client(
    client_id="${FALCON_CLIENT_ID}",
    client_secret="${FALCON_CLIENT_SECRET}",
    cloud_name="${FALCON_CLOUD_NAME}",
)

Whilst this is somewhat an antipattern in the example code provided, it does
allow for the examples to be configured with values provided as environment variables.
It may also be advantageous for some scripting implementations, where the user may wish
to configure Caracara dynamically or interactively by using environment variables.
"""
import logging
import os
import re


class VariableInterpolator:  # pylint: disable=too-few-public-methods
    """Variable interpolation class.

    We define this as a class rather than a standalone function so that we only have to
    compile the regex pattern once. This gives a (very slight) performance gain when performing
    interpolation against a long list of variables.
    """

    def __init__(self):
        """Configure the interpolation regex pattern.

        The pattern will search for all instances of ${text} in a string, but
        will not match on $${text}. This allows a double dollar sign to act as
        an escape character.
        The pattern will return just the inner text (without the $ or {}).
        """
        self._pattern = re.compile(r'(?<!\$)\$\{(\w+)\}')

        self.logger = logging.getLogger(__name__)
        self.logger.debug("Instantiating a Variable Interpolator")

    def interpolate(self, input_string: str) -> str:
        """Take an input string and swap interpolation strings with environment variables.

        Since the regex pattern only returns the inner text, we need to re-add the
        $ and {} to the matched string to find the string to replace.
        This allows us to look up the inner text as an environment variable name, and still
        have the full interpolation string for replacement purposes.
        """
        if not input_string:
            return None

        matches = self._pattern.findall(input_string)
        output_string = input_string
        for match in matches:
            """
            text -> ${text}
            Triple curly braces are used as we need a double curly brace for the
            actual curly brace character, plus a third curly brace for the
            f-string variable name.
            """
            interpolation_str = f'${{{match}}}'
            self.logger.debug("Interpolating the environment variable: %s", match)
            output_string = output_string.replace(
                interpolation_str,
                os.environ.get(match, match)
            )

        # Unescape by replacing double dollar signs with single dollar signs
        output_string = output_string.replace("$$", "$")

        return output_string



"""
Falcon Host module.

Exposes functions to get host data and perform actions on hosts, such as network containment.
"""
__all__ = [
    'CustomIoaApiModule',
    'FlightControlApiModule',
    'HostsApiModule',
    'PreventionPoliciesApiModule',
    'ResponsePoliciesApiModule',
    'RTRApiModule',
    'MODULE_FILTER_ATTRIBUTES',
]

from caracara.modules.custom_ioa import CustomIoaApiModule
from caracara.modules.flight_control import FlightControlApiModule
from caracara.modules.hosts import (
    FILTER_ATTRIBUTES as hosts_filter_attributes,
    HostsApiModule,
)
from caracara.modules.prevention_policies import PreventionPoliciesApiModule
from caracara.modules.response_policies import ResponsePoliciesApiModule
from caracara.modules.rtr import (
    FILTER_ATTRIBUTES as rtr_filter_attributes,
    RTRApiModule,
)

# Build up a list with all filter attributes from the includes modules.
# This makes makes for much easier importing when setting up Falcon Filters.
MODULE_FILTER_ATTRIBUTES = [
    *hosts_filter_attributes,
    *rtr_filter_attributes,
]



"""
Caracara: Response Policies.

This file contains platform-specific templates representing standard, blank Response policies.
"""
from caracara.common.policy_wrapper import (
    ChangeablePolicySetting,
    Policy,
    PolicySettingGroup,
    SETTINGS_TYPE_MAP,
)


COMMAND_TEMPLATES = {
    "RealTimeFunctionality": {
        "description": (
            "Allow those with Real Time Responder roles to remotely connect to hosts. "
            "Required for all RTR commands and scripts and for partner software update policies."
        ),
        "name": "Real Time Response",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "CustomScripts": {
        "description": (
            "Allows those with RTR Active Responder and RTR Administrator roles to run custom "
            "scripts"
        ),
        "name": "Custom Scripts",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "GetCommand": {
        "description": "Extract files from a remote host via the CrowdStrike cloud",
        "name": "get",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "PutCommand": {
        "description": (
            "Send files to a remote host via the CrowdStrike cloud. Required for partner software "
            "update policies"
        ),
        "name": "put",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "MemDumpCommand": {
        "description": "Dump process memory of a remote host",
        "name": "memdump",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "XMemDumpCommand": {
        "description": "Dump the complete memory of a remote host",
        "name": "xmemdump",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "ExecCommand": {
        "description": "Run any executable on the remote host",
        "name": "run",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "PutAndRunCommand": {
        "description": (
            "Send files to a secure directory on a remote host via the CrowdStrike cloud "
            "and execute them with a single command"
        ),
        "name": "put-and-run",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
}

OPTION_PLATFORM_MAPPING = {
    "Linux": {
        "Enable/Disable": [
            "RealTimeFunctionality",
        ],
        "Custom scripts": [
            "CustomScripts"
        ],
        "High risk commands": [
            "GetCommand",
            "PutCommand",
            "ExecCommand",
        ],
    },
    "Mac": {
        "Enable/Disable": [
            "RealTimeFunctionality",
        ],
        "Custom scripts": [
            "CustomScripts"
        ],
        "High risk commands": [
            "GetCommand",
            "PutCommand",
            "ExecCommand",
            "PutAndRunCommand",
        ]
    },
    "Windows": {
        "Enable/Disable": [
            "RealTimeFunctionality",
        ],
        "Custom scripts": [
            "CustomScripts"
        ],
        "High risk commands": [
            "GetCommand",
            "PutCommand",
            "MemDumpCommand",
            "XMemDumpCommand",
            "ExecCommand",
            "PutAndRunCommand",
        ]
    },
}


def generate_response_template(platform_name: str) -> Policy:
    """
    Generate a blank response policy object.

    This policy contains all settings in their default disabled states, ready for
    user or developer modification.
    """
    if platform_name not in OPTION_PLATFORM_MAPPING:
        raise Exception(f"The platform name {platform_name} is not valid")

    policy = Policy(style="response")
    policy.name = "Example Response Policy"
    policy.description = "Response Policy generated from a template"
    policy.platform_name = platform_name

    for group_name, group_settings in OPTION_PLATFORM_MAPPING[platform_name].items():
        group = PolicySettingGroup()
        group.name = group_name

        for setting_id in group_settings:
            setting_data = COMMAND_TEMPLATES[setting_id]
            setting_type = setting_data['setting_type']
            setting_template: ChangeablePolicySetting = SETTINGS_TYPE_MAP[setting_type]
            loadable_setting_data = {
                **COMMAND_TEMPLATES[setting_id],
                "id": setting_id,
            }
            setting = setting_template(data_dict=loadable_setting_data)
            group.settings.append(setting)

        policy.settings_groups.append(group)

    return policy



"""RTR-specific FQL filters."""
from caracara.modules.rtr.constants import RTR_COMMANDS
from caracara.filters.fql import FalconFilterAttribute


class RtrBaseCommandFilterAttribute(FalconFilterAttribute):
    """Filter by Real Time Response (RTR) base command."""

    name = "Command"
    fql = "base_command"
    options = list(RTR_COMMANDS.keys())
    restrict = True
    types = [str]


FILTER_ATTRIBUTES = [
    RtrBaseCommandFilterAttribute,
]



"""Caracara Hosts Module: host containment functions.

In order to avoid the main hosts.py file getting too unwieldly, this file contains
the implementations of the host containment functions.
"""
# Disable the protected access rule because this file is an extension of the class in hosts.py.
# pylint: disable=protected-access
from __future__ import annotations
from typing import (
    Dict,
    TYPE_CHECKING,
)

from caracara.common.exceptions import MustProvideFilter
from caracara.filters import FalconFilter
from caracara.filters.decorators import filter_string

if TYPE_CHECKING:
    from caracara.modules.hosts import HostsApiModule


@filter_string
def contain(
    self: HostsApiModule,
    filters: FalconFilter or str = None,
) -> Dict:
    """Contain a host or list of hosts within your Falcon tenant.

    Arguments
    ---------
    filters: FalconFilter or str, optional
        Filters to apply to the device search.

    Returns
    -------
    dict: A dictionary containing details for the device containment result.
    """
    if not filters:
        raise MustProvideFilter

    return self._perform_action(
        action_name="contain",
        device_ids=self.get_device_ids(filters),
    )["resources"]


@filter_string
def release(
    self: HostsApiModule,
    filters: FalconFilter or str = None,
) -> Dict:
    """Lift containment for a host or list of hosts within your Falcon tenant.

    Arguments
    ---------
    filters: FalconFilter or str, optional
        Filters to apply to the device search.

    Returns
    -------
    dict: A dictionary containing details for the device containment result.
    """
    if not filters:
        raise MustProvideFilter

    return self._perform_action(
        action_name="lift_containment",
        device_ids=self.get_device_ids(filters),
    )["resources"]



"""Caracara Custom Indicator of Attack (IOA) module."""
__all__ = [
    'CustomIoaApiModule',
    'CustomIoaRule',
    'IoaRuleGroup',
]

from caracara.modules.custom_ioa.custom_ioa import CustomIoaApiModule
from caracara.modules.custom_ioa.rules import CustomIoaRule, IoaRuleGroup



"""Caracara Hosts Module: host group functions.

In order to avoid the main hosts.py file getting too unwieldly, this file contains
the implementations of all functions associated with host groups.
"""
# Disable the protected access rule because this file is an extension of the class in hosts.py.
# pylint: disable=protected-access
from __future__ import annotations
from functools import partial
from typing import (
    Dict,
    List,
    TYPE_CHECKING,
)

from caracara.common.batching import batch_get_data
from caracara.common.constants import HOST_GROUP_SCROLL_BATCH_SIZE
from caracara.common.exceptions import (
    GenericAPIError,
    HostGroupNotFound,
    MissingArgument,
    MissingArguments,
    MustProvideFilterOrID,
)
from caracara.common.pagination import all_pages_numbered_offset_parallel
from caracara.filters import FalconFilter
from caracara.filters.decorators import filter_string

if TYPE_CHECKING:
    from caracara.modules.hosts import HostsApiModule


def _create_host_group(
    self: HostsApiModule,
    group_name: str = None,
    description: str = None,
    group_type: str = "static",
    assignment_rule: str = None,
) -> Dict[str, Dict]:
    """Create the requested host group within the tenant."""
    if not group_name:
        raise MissingArgument("group_name")
    if not group_type:
        self.logger.info("Group type not specified for creation, defaulting to static.")

    returned = self.host_group_api.create_host_groups(
        name=group_name,
        description=description,
        group_type=group_type,
        assignment_rule=assignment_rule
    )["body"]
    if returned["errors"]:
        raise GenericAPIError(returned["errors"])

    return returned["resources"]


def _perform_group_action(
    self: HostsApiModule,
    action_name: str,
    group_ids: List[str],
    device_ids: List[str],
) -> Dict:
    """Perform the specified action against the host group."""
    returned = self.host_group_api.perform_group_action(
        ids=group_ids,
        action_name=action_name,
        filter=f"(device_id:{device_ids})",
    )["body"]
    if returned["errors"]:
        raise GenericAPIError(returned["errors"])

    return returned


@filter_string
def add_to_group(
    self: HostsApiModule,
    filters: FalconFilter or str = None,
    group_ids: List[str] or str = None,
    device_filters: FalconFilter or str = None,
    device_ids: List[str] or str = None,
) -> Dict:
    """Add a host or list of hosts to a host group within your Falcon tenant.

    Arguments
    ---------
    filters: FalconFilter or str, optional
        Group filter to apply to the host group search. Not required if group_ids are provided.
    group_ids: List[str] or str, optional
        List of host group IDs to update. Comma delimited strings are converted.
        Not required if a group filter is provided. Takes precedence over provided filters.
    device_filters: FalconFilter or str, optional
        Filters to apply to the device search. Not required if device_ids are provided.
    device_ids: List[str] or str, optional
        List of device IDs to add to the host group. Comma delimited strings are converted.
        Not required if a device filter is provided. Takes precedence over device filters.

    Returns
    -------
    dict: A dictionary containing details for the host group update result.
    """
    if isinstance(group_ids, str):
        group_ids = group_ids.split(",")

    if not group_ids and not filters:
        raise MustProvideFilterOrID

    if isinstance(device_ids, str):
        device_ids = device_ids.split(",")

    if not device_ids and not device_filters:
        raise MustProvideFilterOrID

    return self._perform_group_action(
        action_name="add-hosts",
        group_ids=group_ids if group_ids else self.get_group_ids(filters),
        device_ids=device_ids if device_ids else self.get_device_ids(device_filters),
    )["resources"]


def create_group(
    self: HostsApiModule,
    group_name: str = None,
    description: str = None,
    group_type: str = "static",
    assignment_rule: str = None,
) -> Dict:
    """Create a group based upon the provided arguments within your Falcon tenant.

    Arguments
    ---------
    assignment_rule: str, optional - Assignment rule to apply when creating the host group.
    description: str, optional - Description to use for the created host group.
    group_name: str, required - Name of the host group to create.
    group_type: str, optional - Type of host group to create. `dynamic` or `static`.
        Defaults to `static`.

    Returns
    -------
    dict: A dictionary containing details for the host group creation result.
    """
    return self._create_host_group(
        group_name,
        description,
        group_type,
        assignment_rule
    )


def delete_group(
    self: HostsApiModule,
    group_ids: List[str] = None,
    filters: FalconFilter or str = None,
) -> Dict:
    """Delete a host group from within your Falcon tenant.

    Arguments
    ---------
    filters: FalconFilter or str, optional
        Group filter to apply to the host group search. Not required if group_ids are provided.
    group_ids: List[str] or str, optional
        List of host group IDs to update. Comma delimited strings are converted.
        Not required if a group filter is provided. Takes precedence over provided filters.

    Returns
    -------
    dict: A dictionary containing details for the host group update result.
    """
    if isinstance(group_ids, str):
        group_ids = group_ids.split(",")

    if not group_ids and not filters:
        raise MustProvideFilterOrID

    returned = self.host_group_api.delete_host_groups(
        ids=group_ids if group_ids else self.get_group_ids(filters)
        )["body"]
    if returned["errors"]:
        raise GenericAPIError(returned["errors"])

    return returned["body"]["resources"]


@filter_string
def describe_group_member_ids(
    self: HostsApiModule,
    filters: FalconFilter or str = None,
) -> Dict[str, List[str]]:
    """Return a dictionary with member IDs for all host groups matching the provided filter.

    Arguments
    ---------
    filters: FalconFilter or str, optional
        Filters to apply to the host group search.

    Returns
    -------
    dict: A dictionary containing details for every host group discovered.
    """
    self.logger.info("Describing host group members according to the filter string %s", filters)
    group_ids = self.get_group_ids(filters)
    group_members = {}
    for group_id in group_ids:
        group_members[group_id] = self.get_group_member_ids(group_id)

    return group_members


@filter_string
def describe_group_members(
    self: HostsApiModule,
    filters: FalconFilter or str = None,
) -> Dict:
    """Return a dictionary with member details for all host groups matching the provided filter.

    Arguments
    ---------
    filters: FalconFilter or str, optional
        Filters to apply to the host group search.

    Returns
    -------
    dict: A dictionary containing details for every host group discovered.
    """
    self.logger.info("Describing host group members according to the filter string %s", filters)
    return self.get_group_members(filters)


@filter_string
def describe_groups(
    self: HostsApiModule,
    filters: FalconFilter or str = None,
) -> Dict[str, Dict]:
    """Return a dictionary containing detail for every host group matching the provided filter.

    Arguments
    ---------
    filters: FalconFilter or str, optional
        Filters to apply to the host group search.

    Returns
    -------
    dict: A dictionary containing details for every host group discovered.
    """
    self.logger.info("Describing host groups according to the filter string %s", filters)
    group_ids = self.get_group_ids(filters)
    group_data = batch_get_data(group_ids, self.host_group_api.get_host_groups)

    return group_data


@filter_string
def get_group_ids(
    self: HostsApiModule,
    filters: FalconFilter or str = None,
) -> List[str]:
    """Return a list of IDs (string) for every host group within your Falcon tenant.

    Arguments
    ---------
    filters: FalconFilter or str, optional
        Filters to apply to the host group search.

    Returns
    -------
    List[str]: A list of all host group IDs discovered.
    """
    self.logger.info("Searching for host group IDs using the filter string %s", filters)
    func = partial(self.host_group_api.query_host_groups, filter=filters)
    id_list: List[str] = all_pages_numbered_offset_parallel(
        func=func,
        logger=self.logger,
        limit=HOST_GROUP_SCROLL_BATCH_SIZE
    )
    if not id_list:
        raise HostGroupNotFound

    return id_list


def get_group_member_ids(
    self: HostsApiModule,
    group_id: str = None,
) -> List[str]:
    """Return a list of IDs (string) for every host group member for the specified host group.

    Arguments
    ---------
    group_id: str, required
        ID of the host group to return members for.

    Returns
    -------
    List[str]: A list of all host group member IDs discovered.
    """
    self.logger.info("Searching for host group members using the group ID %s", group_id)
    func = partial(self.host_group_api.query_group_members, id=group_id)
    id_list: List[str] = all_pages_numbered_offset_parallel(
        func=func,
        logger=self.logger,
        limit=HOST_GROUP_SCROLL_BATCH_SIZE,
    )

    return id_list


@filter_string
def get_group_members(
    self: HostsApiModule,
    filters: FalconFilter or str = None,
) -> Dict[str, Dict]:
    """Return a dictionary containing every host group and their members.

    Arguments
    ---------
    filters: FalconFilter or str, optional
        Filters to apply to the host group member search.

    Returns
    -------
    dict: A dictionary of all host groups and their members discovered.
    """
    groups = self.describe_groups()
    func = partial(self.host_group_api.query_combined_group_members, filter=filters)
    device_data = all_pages_numbered_offset_parallel(func, self.logger)
    all_group_data: Dict[str, Dict] = {}
    for group_id, group_data in groups.items():
        all_group_data[group_id] = group_data
        all_group_data[group_id]['devices'] = []

        for device in device_data:
            if 'groups' not in device:
                continue
            for group_identifier in device['groups']:
                if group_identifier == group_id:
                    all_group_data[group_identifier]['devices'].append(device)

    return all_group_data


def group(
    self: HostsApiModule,
    group_name: str = None,
    description: str = None,
    assignment_rule: str = None,
    device_ids: List[str] = None,
) -> Dict:
    """Group a list of device IDs into a newly created host group within the tenant.

    Arguments
    ---------
    assignment_rule: str, optional - Assignment rule to apply when creating the host group.
    description: str, optional - Description to use for the created host group.
    group_name: str, required - Name of the host group to create.
    device_ids: List[str], required - List of device IDs to add to the newly created group.

    Returns
    -------
    dict: A dictionary containing details for the host group creation result.
    """
    if not description:
        description = "Grouped collection of hosts"
    if not device_ids:
        raise MissingArgument("device_ids")
    new_group = self._create_host_group(
        group_name,
        description,
        "static",
        assignment_rule,
    )[0]["id"]
    return self.add_to_group(group_ids=new_group, device_ids=device_ids)["resources"]


@filter_string
def remove_from_group(
    self: HostsApiModule,
    filters: FalconFilter or str = None,
    group_ids: List[str] or str = None,
    device_filters: FalconFilter or str = None,
    device_ids: List[str] or str = None,
) -> Dict:
    """Remove a host or list of hosts to a host group within your Falcon tenant.

    Arguments
    ---------
    filters: FalconFilter or str, optional
        Group filter to apply to the host group search. Not required if group_ids are provided.
    group_ids: List[str] or str, optional
        List of host group IDs to update. Comma delimited strings are converted.
        Not required if a group filter is provided. Takes precedence over provided filters.
    device_filters: FalconFilter or str, optional
        Filters to apply to the device search. Not required if device_ids are provided.
    device_ids: List[str] or str, optional
        List of device IDs to add to the host group. Comma delimited strings are converted.
        Not required if a device filter is provided. Takes precedence over device filters.

    Returns
    -------
    dict: A dictionary containing details for the host group update result.
    """
    if isinstance(group_ids, str):
        group_ids = group_ids.split(",")

    if not group_ids and not filters:
        raise MustProvideFilterOrID

    if isinstance(device_ids, str):
        device_ids = device_ids.split(",")

    if not device_ids and not device_filters:
        raise MustProvideFilterOrID

    return self._perform_group_action(
        action_name="remove-hosts",
        group_ids=group_ids if group_ids else self.get_group_ids(filters),
        device_ids=device_ids if device_ids else self.get_device_ids(device_filters),
    )["resources"]


def ungroup(
    self: HostsApiModule,
    group_ids: List[str] or str = None,
    remove_groups: bool = False,
) -> Dict[str, Dict]:
    """Remove all members from host groups and then remove the group (if specified).

    Arguments
    ---------
    group_ids: List[str] or str, required - List of host group IDs to ungroup.
    remove_groups: bool, optional - Should host groups be removed after ungrouping.

    Returns
    -------
    dict: A dictionary containing details for the host group update results.
    """
    if not group_ids:
        raise MissingArgument("group_ids")
    returned = {}
    if isinstance(group_ids, str):
        group_ids = group_ids.split(",")
    for group_id in group_ids:
        members = self.get_group_member_ids(group_id)
        returned[group_id] = {}
        returned[group_id]["removed"] = False
        if members:
            returned[group_id]["result"] = self.remove_from_group(
                group_ids=group_id,
                device_ids=members,
            )
            self.logger.info(
                "Removed %d members from group %s",
                len(returned[group_id]["result"]),
                group_id
            )
        if remove_groups:
            self.logger.info("Remove group %s", group_id)
            delete_result = self.delete_group(group_ids=group_id)
            if delete_result:
                returned[group_id]["removed"] = True

    return returned["resources"]


def update_group(
    self: HostsApiModule,
    group_id: str = None,
    group_name: str = None,
    group_description: str = None,
    assignment_rule: str = None,
 ) -> Dict:
    """Update the name, assignment rule or description for a Host Group wihtin your tenant.

    Arguments
    ---------
    assignment_rule: str, optional - Assignment rule to apply when creating the host group.
    description: str, optional - Description to use for the created host group.
    group_name: str, optional - Name of the host group to create.
    group_id: str, required - ID of the host group to be updated.

    Returns
    -------
    dict: A dictionary containing details for the host group creation result.
    """
    if not group_id:
        raise MissingArgument("group_id")
    if not group_name and not assignment_rule and not group_description:
        raise MissingArguments(["group_name", "group_description", "assignment_rule"])

    return self.host_group_api.update_host_groups(
        id=group_id,
        assignment_rule=assignment_rule,
        description=group_description,
        name=group_name
    )["body"]["resources"]



"""
Caracara: Prevention Policies.

This file contains platform-specific templates representing standard, blank Prevention policies.
"""
from caracara.common.policy_wrapper import (
    ChangeablePolicySetting,
    Policy,
    PolicySettingGroup,
    SETTINGS_TYPE_MAP,
)


COMMAND_TEMPLATES = {
    "AdditionalUserModeData":
    {
        "description": (
            "Allows the sensor to get more data from a user-mode component it loads into all "
            "eligible processes, which augments online machine learning and turns on additional "
            "detections. Recommend testing with critical applications before full deployment."
        ),
        "name": "Additional User Mode Data",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "AdwarePUP": {
        "description": (
            "Use cloud-based machine learning informed by global analysis of executables to "
            "detect and prevent adware and potentially unwanted programs (PUP) for your online "
            "hosts."
        ),
        "name": "Adware & PUP",
        "setting_type": "mlslider",
        "value": {
            "detection": "DISABLED",
            "prevention": "DISABLED",
        },
    },
    "ApplicationExploitationActivity": {
        "description": (
            "Creation of a process, such as a command prompt, from an exploited browser or "
            "browser flash plugin was blocked."
        ),
        "name": "Application Exploitation Activity",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "BackupDeletion": {
        "description": "Deletion of backups often indicative of ransomware activity.",
        "name": "Backup Deletion",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "ChopperWebshell": {
        "description": (
            "Execution of a command shell was blocked and is indicative of the system "
            "hosting a Chopper web page."
        ),
        "name": "Chopper Webshell",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "CloudAntiMalware": {
        "description": (
            "Use cloud-based machine learning informed by global analysis of executables to "
            "detect and prevent known malware for your online hosts."
        ),
        "name": "Cloud Anti-malware",
        "setting_type": "mlslider",
        "value": {
            "detection": "DISABLED",
            "prevention": "DISABLED",
        },
    },
    "CredentialDumping": {
        "description": "Kill suspicious processes determined to be stealing logins and passwords.",
        "name": "Credential Dumping",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "Cryptowall": {
        "description": "A process associated with Cryptowall was blocked.",
        "name": "Cryptowall",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "CustomBlacklisting": {
        "description": (
            "Block processes matching hashes that you add to IOC Management with the action "
            "set to \"Block\" or \"Block, hide detection\"."
        ),
        "name": "Custom Blocking",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "DetectOnWrite": {
        "description": (
            "Use machine learning to analyze suspicious files when they're written to disk. "
            "To adjust detection sensitivity, change Anti-malware Detection levels in Sensor "
            "Machine Learning and Cloud Machine Learning."
        ),
        "name": "Detect on Write",
        "setting_type": "mlslider",
        "value": {
            "detection": "DISABLED",
            "prevention": "DISABLED",
        },
    },
    "DriveByDownload": {
        "description": (
            "A suspicious file written by a browser attempted to execute and was blocked."
        ),
        "name": "Drive-by Download",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "EmpyreBackdoor": {
        "description": "A process with behaviors indicative of the Empyre Backdoor was terminated.",
        "name": "Empyre Backdoor",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "EndUserNotifications": {
        "description": {
            "Mac": (
                "Show a pop-up notification to the end user when the Falcon sensor blocks, kills, "
                "or quarantines. See these messages in Console.app by searching for Process: "
                "Falcon Notifications."
            ),
            "Windows": (
                "Show a pop-up notification to the end user when the Falcon sensor blocks, "
                "kills, or quarantines. These messages also show up in the Windows Event Viewer "
                "under Applications and Service Logs."
            ),
        },
        "name": "Notify End Users",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "EngineProtectionV2": {
        "description": (
            "Provides visibility into malicious System Management Automation engine usage by "
            "any application. Requires Interpreter-Only."
        ),
        "name": "Engine (Full Visibility)",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "FileEncryption": {
        "description": (
            "A process that created a file with a known ransomware extension was terminated."
        ),
        "name": "File Encryption",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "FileSystemAccess": {
        "description": (
            "A process associated with a high volume of file system operations typical of "
            "ransomware behavior was terminated."
        ),
        "name": "File System Access",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "FilesystemVisibility": {
        "description": (
            "Allows the sensor to monitor filesystem activity for additional telemetry "
            "and improved detections."
        ),
        "name": "Filesystem Visibility",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "FirmwareAnalysisExtraction": {
        "description": (
            "Provides visibility into BIOS. Detects suspicious and unexpected images. Recommend "
            "testing to monitor system startup performance before full deployment."
        ),
        "name": "BIOS Deep Visibility",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "FirmwareDeepVisibility": {
        "description": "Provides visibility into BIOS. Detects suspicious and unexpected images.",
        "name": "BIOS Deep Visibility",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "FirmwareStandardVisibility": {
        "description": (
            "Provides visibility into BIOS configurations to detect vulnerable settings. For "
            "Big Sur, both kernel extension approval and reboot are required."
        ),
        "name": "BIOS Standard Visibility",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "ForceASLR": {
        "description": (
            "An Address Space Layout Randomization (ASLR) bypass attempt was detected and "
            "blocked. This may have been part of an attempted exploit."
        ),
        "name": "Force ASLR",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "ForceDEP": {
        "description": (
            "A process that had Force Data Execution Prevention (Force DEP) applied tried "
            "to execute non-executable memory and was blocked."
        ),
        "name": "Force DEP",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "HardwareEnhancedExploitDetection": {
        "description": (
            "Provides additional visibility into application exploits by using CPU hardware "
            "features that detect suspicious control flows. Available only for hosts running "
            "Windows 10 (RS4) or Windows Server 2016 Version 1803 or later and Skylake or "
            "later CPU."
        ),
        "name": "Hardware-Enhanced Exploit Detection",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "HashCollector": {
        "description": "An attempt to dump a user's hashed password was blocked.",
        "name": "Hash Collector",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "HeapSprayPreallocation": {
        "description": (
            "A heap spray attempt was detected and blocked. This may have been part of an "
            "attempted exploit."
        ),
        "name": "Heap Spray Preallocation",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "HTTPDetections": {
        "description": (
            "Allows the sensor to monitor unencrypted HTTP traffic and certain encrypted HTTPS "
            "traffic on the sensor for malicious patterns and generate detection events on "
            "non-Server systems."
        ),
        "name": "HTTP Detections",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "IntelPrevention": {
        "description": (
            "Block processes that CrowdStrike Intelligence analysts classify as malicious. "
            "These are focused on static hash-based IOCs."
        ),
        "name": "Intelligence-Sourced Threats",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "InterpreterProtection": {
        "description": (
            "Provides visibility into malicious PowerShell interpreter usage. For hosts running "
            "Windows 10, Script-Based Execution Monitoring may be used instead."
        ),
        "name": "Interpreter-Only",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "JavaScriptViaRundll32": {
        "description": "JavaScript executing from a command line via rundll32.exe was prevented.",
        "name": "JavaScript Execution Via Rundll32",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "KcPasswordDecoded": {
        "description": (
            "An attempt to recover a plaintext password via the kcpassword file was blocked."
        ),
        "name": "KcPassword Decoded",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "Locky": {
        "description": "A process determined to be associated with Locky was blocked.",
        "name": "Locky",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "MaliciousPowershell": {
        "description": (
            "Block execution of scripts and commands that CrowdStrike analysts classify as "
            "suspicious. Requires Interpreter-Only and/or Script-Based Execution Monitoring."
        ),
        "name": "Suspicious Scripts and Commands",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "NetworkVisibility": {
        "description": (
            "Allows the sensor to monitor network activity for additional telemetry and "
            "improved detections."
        ),
        "name": "Network Visibility",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "NextGenAV": {
        "description": (
            "Quarantine executable files after they're prevented by NGAV. When this is "
            "enabled, we recommend setting anti-malware prevention levels to Moderate "
            "and not using other antivirus solutions."
        ),
        "name": "Quarantine",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "NullPageAllocation": {
        "description": (
            "Allocating memory to the NULL (0) memory page was detected and blocked. This may "
            "have been part of an attempted exploit."
        ),
        "name": "NULL Page Allocation",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "OnSensorMLAdwarePUPSlider": {
        "description": (
            "For offline and online hosts, use sensor-based machine learning to identify and "
            "analyze unknown executables as they run to detect and prevent adware and "
            "potentially unwanted programs (PUP)."
        ),
        "name": "Adware & PUP",
        "setting_type": "mlslider",
        "value": {
            "detection": "DISABLED",
            "prevention": "DISABLED",
        },
    },
    "OnSensorMLSlider": {
        "description": (
            "For offline and online hosts, use sensor-based machine learning to identify and "
            "analyze unknown executables as they run to detect and prevent malware."
        ),
        "name": "Sensor Anti-malware",
        "setting_type": "mlslider",
        "value": {
            "detection": "DISABLED",
            "prevention": "DISABLED",
        },
    },
    "PreventSuspiciousProcesses": {
        "description": (
            "Block processes that CrowdStrike analysts classify as suspicious. These are focused "
            "on dynamic IOAs, such as malware, exploits and other threats."
        ),
        "name": "Suspicious Processes",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "ProcessHollowing": {
        "description": "Kill processes that unexpectedly injected code into another process.",
        "name": "Code Injection",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "QuarantineOnWrite": {
        "description": (
            "Use machine learning to quarantine suspicious files when they're written to disk. To "
            "adjust quarantine sensitivity, change Anti-malware Prevention levels in Sensor "
            "Machine Learning and Cloud Machine Learning."
        ),
        "name": "Quarantine on Write",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "RedactHTTPDetectionDetails": {
        "description": (
            "Remove certain information from HTTP Detection events, including URL, raw HTTP "
            "header and POST bodies if they were present. This does not affect the generation "
            "of HTTP Detections, only additional details that would be included and may include "
            "personal information (depending on the malware in question). When disabled, the "
            "information is used to improve the response to detection events. Has no effect "
            "unless HTTP Detections is also enabled."
        ),
        "name": "Redact HTTP Detection Details",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "AutomatedRemediation": {
        "description": (
            "Perform advanced remediation for IOA detections to kill processes, quarantine "
            "files, remove scheduled tasks, and clear and delete ASEP registry values."
        ),
        "name": "Advanced Remediation",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "ScriptBasedExecutionMonitoring": {
        "description": {
            "Linux": (
                "Provides visibility into suspicious scripts, including shell and "
                "other scripting languages.",
            ),
            "Mac": (
                "Provides visibility into suspicious scripts, including shell and "
                "other scripting languages."
            ),
            "Windows": (
                "For hosts running Windows 10 and Servers 2016 and later, "
                "provides visibility into suspicious scripts and VBA macros in Office "
                "documents. Requires Quarantine & Security Center Registration toggle "
                "to be enabled."
            ),
        },
        "name": "Script-Based Execution Monitoring",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "SEHOverwriteProtection": {
        "description": (
            "Overwriting a Structured Exception Handler (SEH) was detected and may have been "
            "blocked. This may have been part of an attempted exploit."
        ),
        "name": "SEH Overwrite Protection",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "SensorTamperingProtection": {
        "description": (
            "Blocks attempts to tamper with the sensor. If disabled, the sensor still creates "
            "detections for tampering attempts but doesn't block them. Disabling not recommended."
        ),
        "name": "Sensor Tampering Protection",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "SuspiciousKernelDrivers": {
        "description": (
            "Block the loading of kernel drivers that CrowdStrike analysts have identified as "
            "suspicious. Available on Windows 10 and Windows Server 2016 and later."
        ),
        "name": "Suspicious Kernel Drivers",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "SuspiciousRegistryOperations": {
        "description": (
            "Block registry operations that CrowdStrike analysts classify as suspicious. Focuses "
            "on dynamic IOAs, such as ASEPs and security config changes. The associated process "
            "may be killed."
        ),
        "name": "Suspicious Registry Operations",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "UnknownDetectionRelatedExecutables": {
        "description": (
            "Upload all unknown detection-related executables for advanced "
            "analysis in the cloud."
        ),
        "name": "Unknown Detection-Related Executables",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "UnknownExecutables": {
        "description": "Upload all unknown executables for advanced analysis in the cloud.",
        "name": "Unknown Executables",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "VolumeShadowCopyAudit": {
        "description": (
            "Create an alert when a suspicious process deletes volume shadow copies. "
            "Recommended: Use audit mode with a test group to try allowlisting trusted "
            "software before turning on Protect."
        ),
        "name": "Volume Shadow Copy - Audit",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "VolumeShadowCopyProtect": {
        "description": "Prevent suspicious processes from deleting volume shadow copies.",
        "name": "Volume Shadow Copy - Protect",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "WindowsLogonBypassStickyKeys": {
        "description": (
            "A command line process associated with Windows logon bypass was prevented "
            "from executing."
        ),
        "name": "Windows Logon Bypass (\"Sticky Keys\")",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
    "XPCOMShell": {
        "description": "The execution of an XPCOM shell was blocked.",
        "name": "XPCOM Shell",
        "setting_type": "toggle",
        "value": {
            "enabled": False,
        },
    },
}

OPTION_PLATFORM_MAPPING = {
    "Linux": {
        "User-Mode Capabilities": [
            "UnknownDetectionRelatedExecutables",
            "UnknownExecutables",
        ],
        "Enhanced Visibility": [
            "ScriptBasedExecutionMonitoring",
            "FilesystemVisibility",
            "NetworkVisibility",
        ],
        "Cloud Machine Learning": [
            "CloudAntiMalware",
        ],
        "Sensor Machine Learning": [
            "OnSensorMLSlider",
        ],
        "Execution Blocking": [
            "CustomBlacklisting",
            "PreventSuspiciousProcesses",
        ],
    },
    "Mac": {
        "User-Mode Capabilities": [
            "EndUserNotifications",
            "UnknownDetectionRelatedExecutables",
            "UnknownExecutables",
        ],
        "Enhanced Visibility": [
            "ScriptBasedExecutionMonitoring",
        ],
        "Firmware": [
            "FirmwareStandardVisibility",
            "FirmwareDeepVisibility",
        ],
        "Cloud Machine Learning": [
            "CloudAntiMalware",
            "AdwarePUP",
        ],
        "Sensor Machine Learning": [
            "OnSensorMLSlider",
            "OnSensorMLAdwarePUPSlider",
        ],
        "Quarantine": [
            "NextGenAV",
        ],
        "Execution Blocking": [
            "CustomBlacklisting",
            "PreventSuspiciousProcesses",
            "IntelPrevention",
        ],
        "Unauthorized Remote Access IOAs": [
            "XPCOMShell",
            "ChopperWebshell",
            "EmpyreBackdoor",
        ],
        "Credential Dumping IOAs": [
            "KcPasswordDecoded",
            "HashCollector",
        ],
    },
    "Windows": {
        "User-Mode Capabilities": [
            "EndUserNotifications",
            "UnknownDetectionRelatedExecutables",
            "UnknownExecutables",
            "SensorTamperingProtection",
        ],
        "Enhanced Visibility": [
            "AdditionalUserModeData",
            "InterpreterProtection",
            "EngineProtectionV2",
            "ScriptBasedExecutionMonitoring",
            "HTTPDetections",
            "RedactHTTPDetectionDetails",
            "HardwareEnhancedExploitDetection",
        ],
    },
    "Firmware": [
        "FirmwareAnalysisExtraction",
    ],
    "Cloud Machine Learning": [
        "CloudAntiMalware",
        "AdwarePUP",
    ],
    "Sensor Machine Learning": [
        "OnSensorMLSlider",
    ],
    "On Write": [
        "DetectOnWrite",
        "QuarantineOnWrite",
    ],
    "Quarantine": [
        "NextGenAV",
    ],
    "Execution Blocking": [
        "CustomBlacklisting",
        "PreventSuspiciousProcesses",
        "SuspiciousRegistryOperations",
        "MaliciousPowershell",
        "IntelPrevention",
        "SuspiciousKernelDrivers",
    ],
    "Exploit Mitigation": [
        "ForceASLR",
        "ForceDEP",
        "HeapSprayPreallocation",
        "NullPageAllocation",
        "SEHOverwriteProtection",
    ],
    "Ransomware": [
        "BackupDeletion",
        "Cryptowall",
        "FileEncryption",
        "Locky",
        "FileSystemAccess",
        "VolumeShadowCopyAudit",
        "VolumeShadowCopyProtect",
    ],
    "Exploitation Behavior": [
        "ApplicationExploitationActivity",
        "ChopperWebshell",
        "DriveByDownload",
        "ProcessHollowing",
        "JavaScriptViaRundll32",
    ],
    "Lateral Movement and Credential Access": [
        "WindowsLogonBypassStickyKeys",
        "CredentialDumping",
    ],
    "Remediation": [
        "AutomatedRemediation",
    ],
}


def generate_prevention_template(platform_name: str) -> Policy:
    """
    Generate a blank prevention policy object.

    This policy contains all settings in their default disabled states, ready for
    user or developer modification.
    """
    if platform_name not in OPTION_PLATFORM_MAPPING:
        raise Exception(f"The platform name {platform_name} is not valid")

    policy = Policy(style="prevention")
    policy.name = "Example Prevention Policy"
    policy.description = "Prevention policy generated from a template"
    policy.platform_name = platform_name

    for group_name, group_settings in OPTION_PLATFORM_MAPPING[platform_name].items():
        group = PolicySettingGroup()
        group.name = group_name

        for setting_id in group_settings:
            setting_data = COMMAND_TEMPLATES[setting_id]
            setting_desc = setting_data['description']
            if isinstance(setting_desc, dict):
                setting_desc = setting_desc[platform_name]
            setting_name = setting_data['name']
            setting_type = setting_data['setting_type']
            setting_value = setting_data['value']
            setting_template: ChangeablePolicySetting = SETTINGS_TYPE_MAP[setting_type]
            loadable_setting_data = {
                "description": setting_desc,
                "id": setting_id,
                "name": setting_name,
                "type": setting_type,
                "value": setting_value,
            }
            setting = setting_template(data_dict=loadable_setting_data)
            group.settings.append(setting)

        policy.settings_groups.append(group)

    return policy



r"""Hosts and Host Group APIs.

 ___ ___             __          ___ ___                                                    __
|   Y   .-----.-----|  |_.-----.|   Y   .---.-.-----.---.-.-----.-----.--------.-----.-----|  |_
|.  1   |  _  |__ --|   _|__ --||.      |  _  |     |  _  |  _  |  -__|        |  -__|     |   _|
|.  _   |_____|_____|____|_____||. \_/  |___._|__|__|___._|___  |_____|__|__|__|_____|__|__|____|
|:  |   |                       |:  |   |                 |_____|
|::.|:. |                       |::.|:. |                                   CrowdStrike Caracara
`--- ---'                       `--- ---'

This module handles interactions with the CrowdStrike Falcon Hosts and Host Group APIs.
"""
from functools import partial
from typing import (
    Dict,
    List,
)

from falconpy import (
    Hosts,
    HostGroup,
    OAuth2,
)

from caracara.common.batching import batch_get_data
from caracara.common.exceptions import (
    GenericAPIError,
)
from caracara.common.module import FalconApiModule
from caracara.common.pagination import (
    all_pages_token_offset,
)
from caracara.filters import FalconFilter
from caracara.filters.decorators import filter_string


# Allow importing from other files into this class so that the file does not end up
# being over 1000 lines long.
# pylint: disable=import-outside-toplevel
class HostsApiModule(FalconApiModule):
    """The HostsApiModule represents interactions with the Hosts and Host Group APIs."""

    name = "CrowdStrike Hosts API Module"
    help = "Interact with hosts and host groups within your Falcon tenant."

    def __init__(self, api_authentication: OAuth2):
        """Construct an instance of HostApiModule class."""
        super().__init__(api_authentication)

        self.logger.debug("Configuring the FalconPy Hosts API")
        self.hosts_api = Hosts(auth_object=self.api_authentication)

        self.logger.debug("Configuring the FalconPy Host Group API")
        self.host_group_api = HostGroup(auth_object=self.api_authentication)

    # Import containment functions
    from caracara.modules.hosts._containment import (
        contain,
        release,
    )

    # Import data history functions
    from caracara.modules.hosts._data_history import (
        describe_login_history,
        describe_network_address_history,
        describe_state,
    )

    # Import functions pertaining to host groups
    from caracara.modules.hosts._groups import (
        _create_host_group,
        _perform_group_action,
        add_to_group,
        create_group,
        delete_group,
        describe_group_member_ids,
        describe_group_members,
        describe_groups,
        get_group_ids,
        get_group_member_ids,
        get_group_members,
        group,
        remove_from_group,
        ungroup,
        update_group,
    )

    # Import functions to hide, unhide and describe hidden devices
    from caracara.modules.hosts._hiding import (
        describe_hidden_devices,
        get_hidden_ids,
        hide,
        unhide,
    )

    # Import functions to create and modify device tags
    from caracara.modules.hosts._tagging import (
        _create_tag_list,
        _update_device_tags,
        tag,
        untag,
    )

    # Static methods have to be set within the class
    _create_tag_list = staticmethod(_create_tag_list)

    def _perform_action(self, action_name: str, device_ids: List[str]) -> Dict:
        """Perform the specified action against the list of targets."""
        returned = self.hosts_api.perform_action(ids=device_ids, action_name=action_name)["body"]

        if returned["errors"]:
            raise GenericAPIError(returned["errors"])

        return returned

    @filter_string
    def describe_devices(self, filters: FalconFilter or str = None) -> Dict[str, Dict]:
        """Return a dictionary containing details for every device matching the provided filter.

        Arguments
        ---------
        filters: FalconFilter or str, optional
            Filters to apply to the device search.

        Returns
        -------
        dict: A dictionary containing details for every device discovered.
        """
        self.logger.info("Describing devices according to the filter string %s", filters)
        device_ids = self.get_device_ids(filters)
        device_data = self.get_device_data(device_ids)

        return device_data

    def get_device_data(self, device_ids: List[str]) -> Dict[str, Dict]:
        """Return a dictionary containing details for every device specified by ID.

        You should only use this endpoint if you already have a list of Device IDs / AIDs,
        and want to retreive data about them directly. If you need to get device data via a
        filter, and you do not yet have a list of Device IDs, you should consider using the
        describe_devices() function.

        Arguments
        ---------
        device_ids: [str], required
            A list of Falcon Device IDs.

        Returns
        -------
        dict: A dictionary containing details for every device listed.
        """
        self.logger.info("Obtaining data for %s devices", len(device_ids))
        device_data = batch_get_data(device_ids, self.hosts_api.get_device_details)
        return device_data

    @filter_string
    def get_device_ids(self, filters: FalconFilter or str = None) -> List[str]:
        """Return a list of IDs (string) for every device in your Falcon tenant.

        Arguments
        ---------
        filters: FalconFilter or str, optional
            Filters to apply to the device search.

        Returns
        -------
        List[str]: A list of all device IDs discovered.
        """
        self.logger.info("Searching for device IDs using the filter string %s", filters)
        func = partial(self.hosts_api.query_devices_by_filter_scroll, filter=filters)
        id_list: List[str] = all_pages_token_offset(func=func, logger=self.logger)
        return id_list



"""Caracara HostsApiModule."""
__all__ = [
    'HostsApiModule',
    'FILTER_ATTRIBUTES',
]

from caracara.modules.hosts.hosts import HostsApiModule
from caracara.modules.hosts.hosts_filters import FILTER_ATTRIBUTES



"""Caracara Indicator of Attack (IOA) API module."""
from functools import partial
from time import monotonic
from typing import Dict, List

from falconpy import OAuth2, CustomIOA

from caracara.common.batching import batch_get_data
from caracara.common.constants import DEFAULT_COMMENT
from caracara.common.pagination import all_pages_numbered_offset_parallel
from caracara.filters import FalconFilter
from caracara.filters.decorators import filter_string
from caracara.common.module import FalconApiModule
from caracara.modules.custom_ioa.rules import CustomIoaRule, IoaRuleGroup, RuleType


# instrument falconpy to raise exceptions
def instr(func) -> dict:
    """Instrument FalconPy functions with basic error handling.

    This is an internal function, and therefore developers should not expect this function to
    remain consistent.
    """
    def handle_errors(*args, **kwargs):
        response = func(*args, **kwargs)
        errors = response["body"].get("errors", [])
        if len(errors) == 0:
            return response
        raise Exception(errors)
    return handle_errors


class CustomIoaApiModule(FalconApiModule):
    """Caracara Custom Indicator of Attack (IOA) API module."""

    name = "CrowdStrike Custom IOA Module"
    help = "Interact with Custom IOA Rules and IOA Rule Groups within your Falcon tenant."

    # Don't access directly, always use _get_rule_types_cached
    _rule_type_cache: Dict[str, RuleType]
    _rule_type_cache_time: int = None
    _rule_type_cache_ttl: int = 10 * 60  # 10 minute default cache TTL

    def __init__(self, api_authentication: OAuth2):
        """Create an Custom IOA API object and configure it with a FalconPy OAuth2 object."""
        super().__init__(api_authentication)
        self.custom_ioa_api = CustomIOA(auth_object=api_authentication)

    def _get_rule_types_cached(self, force_update: bool = False) -> Dict[str, RuleType]:
        cur_time = monotonic()
        # If cache has expired
        if self._rule_type_cache_time is None:
            force_update = True
        if force_update or cur_time >= self._rule_type_cache_time + self._rule_type_cache_ttl:
            self._rule_type_cache = self.describe_rule_types()
            self._rule_type_cache_time = cur_time
        return self._rule_type_cache

    def create_rule_group(
        self, group: IoaRuleGroup, comment: str = DEFAULT_COMMENT
    ) -> IoaRuleGroup:
        """Create a new rule group from a manually constructed `IoaRuleGroup`.

        This creates the provided rule group (provided it does not already exist), and also creates
        all the rules within the rule group (added with `IoaRuleGroup.add_rule`).

        Arguments
        ---------
        `group`: `IoaRuleGroup`
            The group to create in the cloud (note this group must not already exist in the cloud,
            i.e. `group.exists_in_cloud() == False`)
        `comment`: `str`
            The comment to associate with this action in the audit log

        Returns
        -------
        `IoaRuleGroup`: The object representing the newly created rule group in the cloud
        """
        if group.exists_in_cloud():
            raise Exception("This group already exists in the cloud!")

        group.validation()

        # Create the group
        group_create = group.dump_create(comment=comment)
        response = instr(self.custom_ioa_api.create_rule_group)(body=group_create)
        new_group = IoaRuleGroup.from_data_dict(
            data_dict=response["body"]["resources"][0], rule_type_map=self._get_rule_types_cached())
        new_group.rules = group.rules
        for rule in new_group.rules:
            rule.group = new_group

        # Update the rules
        new_group = self._create_update_delete_rules(new_group, comment=comment)

        return new_group

    def update_rule_group(
        self, group: IoaRuleGroup, comment: str = DEFAULT_COMMENT
    ) -> IoaRuleGroup:
        """Update a group that already exists in the cloud.

        This will also sync the state of the rules with the cloud, i.e. creating rules that don't
        yet exist, updating existing rules that have changes, and deleting any locally removed
        rules.

        Arguments
        ---------
        `group`: `IoaRuleGroup`
            The group to update.
        `comment`: `str`
            The comment to associate with this action in the audit log

        Returns
        -------
        `IoaRuleGroup`: The updated group.
        """
        if not group.exists_in_cloud():
            raise Exception("This group does not exist in the cloud!")

        group.validation()

        # Update the group
        group_update = group.dump_update(comment=comment)
        response = instr(self.custom_ioa_api.update_rule_group)(body=group_update)
        new_group = IoaRuleGroup.from_data_dict(
            data_dict=response["body"]["resources"][0], rule_type_map=self._get_rule_types_cached())

        # Retain staged rules (as these aren't in the cloud yet, but will be soon)
        new_group.rules = group.rules
        new_group.rules_to_delete = group.rules_to_delete

        # Update the rules
        new_group = self._create_update_delete_rules(new_group, comment=comment)

        return new_group

    def delete_rule_groups(
        self, rule_groups: List[IoaRuleGroup or str], comment: str = DEFAULT_COMMENT
    ):
        """Delete a list of rule groups in the cloud.

        Arguments
        ---------
        `rule_groups`: `List[IoaRuleGroup or str]`
            The list of rule groups as objects or just IDs
        `comment`: `str`
            The comment to associated with this deletion.
        """
        if len(rule_groups) == 0:  # Do nothing if no rule groups to delete
            return

        ids_to_delete = []
        for rule_group in rule_groups:
            if isinstance(rule_group, IoaRuleGroup):
                ids_to_delete.append(rule_group.id_)
            else:
                ids_to_delete.append(rule_group)
        instr(self.custom_ioa_api.delete_rule_groups)(ids=ids_to_delete, comment=comment)

    def _create_update_delete_rules(self, group: IoaRuleGroup, comment: str) -> IoaRuleGroup:
        existing_rules = []
        to_be_created = []
        for rule in group.rules:
            if rule.exists_in_cloud():
                existing_rules.append(rule)
            else:
                to_be_created.append(rule)

        # Create the new rules
        new_rules = []
        for rule in to_be_created:
            resp = instr(self.custom_ioa_api.create_rule)(
                body=rule.dump_create(comment=comment))
            raw_rule = resp["body"]["resources"][0]
            new_rule = CustomIoaRule.from_data_dict(
                raw_rule, rule_type=self._get_rule_types_cached()[raw_rule["ruletype_id"]])
            new_rule.rulegroup_id = group.id_
            new_rules.append(new_rule)

        # Update the existing rules, if there are any
        if len(existing_rules) > 0:
            response = instr(self.custom_ioa_api.update_rules)(body={
                "comment": comment,
                "rule_updates": [rule.dump_update() for rule in existing_rules],
                "rulegroup_version": group.version + 1,
                "rulegroup_id": group.id_,
            })
            raw_group = response["body"]["resources"][0]

            # Create the object representing the updated group
            rule_types = self._get_rule_types_cached()
            new_group = IoaRuleGroup.from_data_dict(data_dict=raw_group, rule_type_map=rule_types)
        else:
            group.rules = new_rules
            new_group = group

        # Delete rules queued for deletion, if any
        if len(group.rules_to_delete) > 0:
            ids_to_delete = [rule.instance_id for rule in group.rules_to_delete]
            instr(self.custom_ioa_api.delete_rules)(
                rule_group_id=group.id_, ids=ids_to_delete, comment=comment)
            # If successful (i.e. no exceptions raised), clear the deletion queue
            group.rules_to_delete = []
        return new_group

    @filter_string
    def describe_rule_groups_raw(self, filters: str or FalconFilter = None) -> Dict[str, dict]:
        """Return all IOA Rule Groups as raw dicts returned from the API, optionally filtered.

        Arguments
        ---------
        `filters`: `FalconFilter` or `str`, optional
            Filters to apply to IOA rule group search

        Returns
        -------
        `Dict[str, IoaRuleGroup]`:
            Dictionary mapping ID to IOA rule group as a raw dict for all the rule groups that match
            the filter.
        """
        self.logger.info("Describing all Falcon IOA Rule Groups matching filter: %s", repr(filters))
        partial_func = partial(
            instr(self.custom_ioa_api.query_rule_groups_full),
            filter=filters,
        )
        resources = all_pages_numbered_offset_parallel(func=partial_func, logger=self.logger)
        self.logger.debug(resources)
        result = {}
        for rule_group in resources:
            result[rule_group["id"]] = rule_group
        return result

    @filter_string
    def describe_rule_groups(self, filters: str or FalconFilter = None) -> Dict[str, IoaRuleGroup]:
        """Return all IOA Rule Groups, optionally filtered.

        Arguments
        ---------
        `filters`: `FalconFilter` or `str`, optional
            Filters to apply to IOA rule group search

        Returns
        -------
        `Dict[str, IoaRuleGroup]`:
            Dictionary mapping ID to IoaRuleGroup for all the rule groups that match the filter.
        """
        raw_groups = self.describe_rule_groups_raw(filters=filters)
        groups = {}
        rule_types = self._get_rule_types_cached()

        for (group_id, raw_group) in raw_groups.items():
            groups[group_id] = IoaRuleGroup.from_data_dict(
                data_dict=raw_group, rule_type_map=rule_types)
        return groups

    def get_rule_types_raw(self, rule_type_ids: List[str]) -> Dict[str, dict]:
        """Get rule types as raw dicts returned by the API, by ID.

        Arguments
        ---------
        `rule_type_ids`: `List[str]`
            List of rule type IDs

        Returns
        -------
        `Dict[str, dict]`:
            Dictionary mapping ID to its associated rule type as a raw dict returned by the api.
        """
        rule_types = batch_get_data(
            rule_type_ids, instr(self.custom_ioa_api.get_rule_types)
        )

        return rule_types

    def get_rule_types(self, rule_type_ids: List[str]) -> Dict[str, RuleType]:
        """Get rule types by ID.

        Arguments
        ---------
        `rule_type_ids`: `List[str]`
            List of rule type IDs

        Returns
        -------
        `Dict[str, RuleType]`:
            Dictionary mapping ID to its associated rule type.
        """
        rule_types = self.get_rule_types_raw(rule_type_ids=rule_type_ids)
        for i, rule_type in rule_types.items():
            rule_types[i] = RuleType.from_data_dict(rule_type)
        return rule_types

    def get_rule_type_ids(self) -> List[str]:
        """Get all rule type IDs.

        Returns
        -------
        `List[str]` list of rule type IDs
        """
        rule_type_ids = all_pages_numbered_offset_parallel(
            instr(self.custom_ioa_api.query_rule_types), logger=self.logger)
        return rule_type_ids

    def describe_rule_types(self) -> Dict[str, RuleType]:
        """Get all rule types.

        Returns
        -------
        `Dict[str, RuleType]`: Dictionary mapping ID to its associated rule type
        """
        rule_type_ids = self.get_rule_type_ids()
        rule_types = self.get_rule_types(rule_type_ids=rule_type_ids)
        return rule_types



"""Caracara meta class.

Returns information about Caracara itself, such as its package version and
a custom user agent string based on Caracara's version.
"""

import pkg_resources

_pkg_version = pkg_resources.get_distribution("caracara").version


def user_agent_string():
    """
    Return the default Caracara user agent.

    This user agent will be sent to CrowdStrike with API requests, including Caracara's version,
    unless the developer overrides this when instantiating a Client object.
    """
    return f"crowdstrike-caracara/{_pkg_version}"
